from abc import abstractmethod


class PartialMixin(object):
    """
    Mixin to be used for API resource objects that have a partial set of fields in their list API, and a full set of
    fields in the get API.
    """

    PARTIAL_ATTR = '_is_partial'

    @property
    def is_partial(self):
        """
        :return: Is it a partial resource?
        :rtype: bool
        """
        return getattr(self, self.PARTIAL_ATTR, True)  # default=True: safer to assume partial resource

    @is_partial.setter
    def is_partial(self, value):
        setattr(self, self.PARTIAL_ATTR, value)

    @abstractmethod
    def refresh(self):
        """
        Fetches the updated resource from the server, and adds/updates object attributes based on it.

        :raises: :py:class:`~utils.error.APIError`: If the api returns an error, the status code is
            anything other than 200/201
        """
        raise NotImplementedError



from __future__ import absolute_import
import enum
from rapyuta_io.utils.object_converter import ObjBase, nested_field


class Organization(ObjBase):
    """
    Organization is the representation of a group of users on the Platform. Every user must be part of at least one
    organization.

    :ivar guid: guid of the Organization
    :vartype guid: str
    :ivar name: Name of the Organization
    :vartype name: str
    :ivar country_id: County ID of the Organization
    :vartype country_id: int
    :ivar country: Country details of the Organization
    :vartype country:   :py:class:`~rapyuta_io.clients.organization.Country`
    :ivar province: Province of the Organization
    :vartype province: str
    :ivar postal_code: Postal Code of the Organization
    :vartype postal_code: str
    :ivar url: URL of the Organization
    :vartype url: str
    :ivar plan_id: Plan ID of the Organization
    :vartype plan_id: int
    :ivar state: The state of the Organization on the Platform
    :vartype state: :py:class:`~rapyuta_io.clients.organization.OrganizationState`
    :ivar creator: GUID of the User that created the Organization
    :vartype creator: str
    :ivar short_guid: short guid of the Organization
    :vartype short_guid: str
    """

    def __init__(self):
        self.guid = None
        self.name = None
        self.country_id = None
        self.country = None
        self.province = None
        self.postal_code = None
        self.url = None
        self.plan_id = None
        self.state = None
        self.creator = None
        self.short_guid = None

    def get_deserialize_map(self):
        return {
            'guid': 'guid',
            'name': 'name',
            'country_id': 'countryID',
            'country': nested_field('country', Country),
            'province': 'province',
            'postal_code': 'postalCode',
            'url': 'url',
            'plan_id': 'planID',
            'state': 'state',
            'creator': 'creator',
            'short_guid': 'shortGUID'
        }

    def get_serialize_map(self):
        return {
            'guid': 'guid',
        }



class OrganizationState(str, enum.Enum):
    """
    Enumeration variables for OrganizationState.

    OrganizationState can be any of the following types \n

    OrganizationState.ACTIVATED ('ACTIVATED') \n
    OrganizationState.SUSPENDED ('SUSPENDED') \n
    OrganizationState.MARKEDSUSPENSION ('MARKEDSUSPENSION') \n

    """
    ACTIVATED = 'ACTIVATED'
    SUSPENDED = 'SUSPENDED'
    MARKEDSUSPENSION = 'MARKEDSUSPENSION'


class Country(ObjBase):
    """
    Country is the representation of country details of an Organization on the platform.

    :ivar id: id of the Country
    :vartype id: int
    :ivar created_at: creation time of the Country details on the platform
    :vartype created_at: str
    :ivar updated_at: updation time of the Country details on the platform
    :vartype updated_at: str
    :ivar deleted_at: deletion time of the Country details on the platform
    :vartype deleted_at: str
    :ivar name: name of the Country
    :vartype name: str
    :ivar code: code of the Country
    :vartype code: str
    """
    def __init__(self):
        self.id = None
        self.created_at = None
        self.updated_at = None
        self.deleted_at = None
        self.name = None
        self.code = None

    def get_deserialize_map(self):
        return {
            'id': 'ID',
            'created_at': 'CreatedAt',
            'updated_at': 'UpdatedAt',
            'deleted_at': 'DeletedAt',
            'name': 'name',
            'code': 'code',
        }

    def get_serialize_map(self):
        pass



# encoding: utf-8
from __future__ import absolute_import

import time
import typing
from enum import Enum

import requests

from rapyuta_io.clients.device import Device, DeviceStatus
from rapyuta_io.clients.model import Command
from rapyuta_io.utils import RestClient
from rapyuta_io.utils.error import ParameterMissingException
from rapyuta_io.utils.rest_client import HttpMethod
from rapyuta_io.utils.settings import DAEMONS_PATH, DEVICE_API_ADD_DEVICE_PATH, DEVICE_API_PATH, \
    DEVICE_COMMAND_API_PATH, DEVICE_SELECTION_API_PATH, PARAMETERS_API_PATH
from rapyuta_io.utils.utils import create_auth_header, get_api_response_data, get_error, prepend_bearer_to_auth_token, \
    validate_list_of_strings


class DeviceArch(str, Enum):
    """
    DeviceArch enumeration represents supported device architectures.
    Device architecture can be any of the below types \n
    DeviceArch.ARM32V7 \n
    DeviceArch.ARM64V8 \n
    DeviceArch.AMD64 \n
    """
    ARM32V7 = 'arm32v7'
    ARM64V8 = 'arm64v8'
    AMD64 = 'amd64'


class DeviceManagerClient:

    def __init__(self, auth_token, project, device_api_host):
        self._device_api_host = device_api_host
        self._auth_token = prepend_bearer_to_auth_token(auth_token)
        self._project = project

    def _add_auth_token_to_devices(self, devices):
        for device in devices:
            setattr(device, '_device_api_host', self._device_api_host)
            setattr(device, '_auth_token', self._auth_token)
            setattr(device, '_project', self._project)

    def _get_device(self, device_id=None, retry_limit=0, device_name=None):
        url = self._device_api_host + DEVICE_API_PATH
        if device_id:
            url = url + device_id

        query = {}
        if device_name is not None:
            query = {"name": device_name}

        headers = create_auth_header(self._auth_token, self._project)
        response = RestClient(url).retry(retry_limit).headers(headers).query_param(query_param=query).execute()
        return get_api_response_data(response)

    @staticmethod
    def _get_specs_cpuarch_query(arch_list):
        args = []
        if DeviceArch.ARM32V7 in arch_list:
            args.append({"operator": "$in", "args": ["cpuarch", "armv7"]})
        if DeviceArch.ARM64V8 in arch_list:
            args.append({"operator": "$in", "args": ["cpuarch", "aarch64"]})
        if DeviceArch.AMD64 in arch_list:
            args.append({"operator": "$eq", "args": ["cpuarch", "x86_64"]})
            args.append({"operator": "$eq", "args": ["cpuarch", "amd64"]})
        return {"operator": "$or", "specs": {"operator": "$or", "args": args}}

    def _device_selection_by_arch(self, arch_list, retry_limit):
        url = self._device_api_host + DEVICE_SELECTION_API_PATH
        headers = create_auth_header(self._auth_token, self._project)
        payload = self._get_specs_cpuarch_query(arch_list)
        response = RestClient(url).method(HttpMethod.POST).retry(retry_limit) \
            .headers(headers).execute(payload=payload)
        return get_api_response_data(response)

    def set_project(self, project):
        self._project = project

    def device_list(self, online_device=False, arch_list=None, retry_limit=0, device_name=None):
        arch_filtered_uuids = set()
        if arch_list:
            for device in self._device_selection_by_arch(arch_list, retry_limit):
                arch_filtered_uuids.add(device['uuid'])

        # TODO(shivam): if arch_list is set there's no need for _get_device all
        device_list = self._get_device(retry_limit=retry_limit, device_name=device_name)
        devices = []
        # todo: add a generic filter like status, name etc
        for device in device_list:
            device = Device._deserialize(device)
            if online_device and device.status != DeviceStatus.ONLINE.value:
                continue
            if arch_list and device.uuid not in arch_filtered_uuids:
                continue
            devices.append(device)
        self._add_auth_token_to_devices(devices)
        return devices

    def get_device(self, device_id, retry_limit):
        device_data = self._get_device(device_id, retry_limit)
        device = Device._deserialize(device_data)
        self._add_auth_token_to_devices([device])
        device.is_partial = False
        return device

    def apply_parameters(self, device_list, tree_names=None, retry_limit=0):
        validate_list_of_strings(device_list, 'device_list')
        if tree_names:
            validate_list_of_strings(tree_names, 'tree_names')

        url = self._device_api_host + PARAMETERS_API_PATH
        headers = create_auth_header(self._auth_token, self._project)
        payload = {'device_list': device_list}
        if tree_names:
            payload['tree_names'] = tree_names
        response = RestClient(url).method(HttpMethod.POST).retry(retry_limit) \
            .headers(headers).execute(payload=payload)
        return get_api_response_data(response)

    def create_device(self, device):
        url = self._device_api_host + DEVICE_API_ADD_DEVICE_PATH
        headers = create_auth_header(self._auth_token, self._project)
        response = RestClient(url).method(HttpMethod.POST).headers(headers).execute(payload=device._serialize())
        return get_api_response_data(response, parse_full=True)

    def delete_device(self, device_id):
        url = self._device_api_host + DEVICE_API_PATH + device_id
        headers = create_auth_header(self._auth_token, self._project)
        return RestClient(url).method(HttpMethod.DELETE).headers(headers).execute()

    def patch_daemons(self, device_id, payload):
        url = self._device_api_host + DEVICE_API_PATH + device_id + DAEMONS_PATH
        headers = create_auth_header(self._auth_token, self._project)
        response = RestClient(url).method(HttpMethod.PATCH).headers(headers).execute(payload=payload)
        return get_api_response_data(response, parse_full=True)

    def execute_command(
            self,
            device_ids: typing.List[str],
            command: Command,
            retry_limit: int = 0,
            retry_interval: int = 10,
            timeout: int = 300,
    ):
        """Execute a command on the specified devices.

        Args:
            device_ids: List of device IDs on which the command should be executed.
            command: Command object to be executed.
            retry_limit: Number of retries in case of API failure.
            retry_interval: Interval between retries.
            timeout: Maximum time to wait for the background command to finish.

        Returns:
            dict: Output of the command execution.

        Raises:
            ValueError: If device_ids is empty.
            TimeoutError: If command execution takes longer than the specified timeout.
            ParameterMissingException: If the command is missing required parameters.
        """
        if not device_ids:
            raise ValueError("device_ids cannot be empty")

        command.validate()
        command.device_ids = device_ids

        url = self._device_api_host + DEVICE_COMMAND_API_PATH
        rc = RestClient(url).method(HttpMethod.POST).headers(
            create_auth_header(self._auth_token, self._project))
        response = rc.retry(retry_limit).execute(payload=command.to_json())
        if response.status_code == requests.codes.BAD_REQUEST:
            raise ParameterMissingException(get_error(response.text))

        execution_result = get_api_response_data(response)

        if not command.run_async:
            return execution_result

        jid = execution_result.get('jid')
        if not jid:
            raise ValueError("job id not found in the response")

        url = self._device_api_host + DEVICE_COMMAND_API_PATH + jid
        query = {"jid": jid, "device_id": device_ids}
        time_elapsed = 0
        wait_interval = retry_interval
        while time_elapsed < timeout:
            response = RestClient(url).method(HttpMethod.GET).headers(
                create_auth_header(self._auth_token, self._project)).query_param(query_param=query).execute()
            if response.status_code == requests.codes.OK:
                result = get_api_response_data(response)
                return result
            time.sleep(wait_interval)
            time_elapsed += wait_interval

        raise TimeoutError(f"command result not available after {timeout} seconds")



from rapyuta_io.clients.project import Project, User
from rapyuta_io.clients.user_group import UserGroup
from rapyuta_io.utils.utils import prepend_bearer_to_auth_token, create_auth_header, get_api_response_data
from rapyuta_io.utils.rest_client import HttpMethod
from rapyuta_io.utils import RestClient
from rapyuta_io.utils.settings import METRICS_API_QUERY_PATH, LIST_METRICS_API_QUERY_PATH, \
    LIST_TAGS_KEY_API_QUERY_PATH, LIST_TAGS_VALUE_API_QUERY_PATH, GET_USER_PATH


class CoreAPIClient:
    def __init__(self, auth_token, project, core_api_host):
        self._core_api_host = core_api_host
        self._auth_token = prepend_bearer_to_auth_token(auth_token)
        self._project = project

    def set_project(self, project):
        self._project = project

    def _add_header_fields(self, obj):
        setattr(obj, '_core_api_host', self._core_api_host)
        setattr(obj, '_auth_token', self._auth_token)
        if type(obj) is not Project:
            setattr(obj, '_project', self._project)

    def _add_auth_token_to_routes(self, routes):
        for route in routes:
            self._add_header_fields(route)

    def query_metrics(self, metrics_query):
        url = self._core_api_host + METRICS_API_QUERY_PATH
        headers = create_auth_header(self._auth_token, self._project)
        payload = metrics_query.serialize()
        response = RestClient(url).method(HttpMethod.POST).headers(headers).execute(payload)
        return get_api_response_data(response, parse_full=False)

    def list_metrics(self, list_metrics_query):
        url = self._core_api_host + LIST_METRICS_API_QUERY_PATH.format(list_metrics_query.entity,
                                                                       list_metrics_query.entity_id)
        headers = create_auth_header(self._auth_token, self._project)
        params = {
            'start_date': list_metrics_query.start_date.isoformat(),
            'end_date': list_metrics_query.end_date.isoformat()
        }
        response = RestClient(url).method(HttpMethod.GET).query_param(params).headers(headers).execute()
        return get_api_response_data(response, parse_full=False)

    def list_tag_keys(self, list_tag_keys_query):
        url = self._core_api_host + LIST_TAGS_KEY_API_QUERY_PATH.format(list_tag_keys_query.entity,
                                                                        list_tag_keys_query.entity_id)
        headers = create_auth_header(self._auth_token, self._project)
        params = {
            'start_date': list_tag_keys_query.start_date.isoformat(),
            'end_date': list_tag_keys_query.end_date.isoformat()
        }
        response = RestClient(url).method(HttpMethod.GET).query_param(params).headers(headers).execute()
        return get_api_response_data(response, parse_full=False)

    def list_tag_values(self, list_tag_values_query):
        url = self._core_api_host + LIST_TAGS_VALUE_API_QUERY_PATH.format(list_tag_values_query.entity,
                                                                          list_tag_values_query.entity_id,
                                                                          list_tag_values_query.tag)
        headers = create_auth_header(self._auth_token, self._project)
        params = {
            'start_date': list_tag_values_query.start_date.isoformat(),
            'end_date': list_tag_values_query.end_date.isoformat()
        }
        response = RestClient(url).method(HttpMethod.GET).query_param(params).headers(headers).execute()
        return get_api_response_data(response, parse_full=False)

    def get_user(self):
        url = self._core_api_host + GET_USER_PATH
        headers = create_auth_header(self._auth_token, self._project)
        response = RestClient(url).method(HttpMethod.GET).headers(headers).execute()
        data = get_api_response_data(response, parse_full=True)
        return User.deserialize(data)

    def get_user_organizations(self):
        url = self._core_api_host + GET_USER_PATH
        headers = create_auth_header(self._auth_token, self._project)
        response = RestClient(url).method(HttpMethod.GET).headers(headers).execute()
        data = get_api_response_data(response, parse_full=True)
        user = User.deserialize(data)
        return user.organizations

    def list_usergroups(self, org_guid):
        url = '{}/api/group/list'.format(self._core_api_host)
        headers = create_auth_header(self._auth_token, self._project)
        headers['organization'] = org_guid
        response = RestClient(url).method(HttpMethod.GET).headers(headers).execute()
        data = get_api_response_data(response, parse_full=True)
        usergroups = []
        for usergroup_data in data:
            usergroup = UserGroup.deserialize(usergroup_data)
            self._add_header_fields(usergroup)
            usergroups.append(usergroup)

        return usergroups

    def get_usergroup(self, org_guid, group_guid):
        url = '{}/api/group/{}/get'.format(self._core_api_host, group_guid)
        headers = create_auth_header(self._auth_token, self._project)
        headers['organization'] = org_guid
        response = RestClient(url).method(HttpMethod.GET).headers(headers).execute()
        data = get_api_response_data(response, parse_full=True)
        usergroup = UserGroup.deserialize(data)
        self._add_header_fields(usergroup)
        return usergroup

    def delete_usergroup(self, org_guid, group_guid):
        url = '{}/api/group/delete'.format(self._core_api_host)
        headers = create_auth_header(self._auth_token, self._project)
        headers['organization'] = org_guid
        payload = {'guid': group_guid}
        response = RestClient(url).method(HttpMethod.DELETE).headers(headers).execute(payload)
        return get_api_response_data(response, parse_full=True)

    def create_usergroup(self, org_guid, usergroup_payload):
        url = '{}/api/group/create'.format(self._core_api_host)
        headers = create_auth_header(self._auth_token, self._project)
        headers['organization'] = org_guid
        response = RestClient(url).method(HttpMethod.POST).headers(headers).execute(usergroup_payload)
        data = get_api_response_data(response, parse_full=True)
        usergroup = UserGroup.deserialize(data)
        self._add_header_fields(usergroup)
        return usergroup

    def update_usergroup(self, org_guid, group_guid, usergroup_payload):
        url = '{}/api/group/{}/update'.format(self._core_api_host, group_guid)
        headers = create_auth_header(self._auth_token, self._project)
        headers['organization'] = org_guid
        response = RestClient(url).method(HttpMethod.PUT).headers(headers).execute(usergroup_payload)
        data = get_api_response_data(response, parse_full=True)
        usergroup = UserGroup.deserialize(data)
        self._add_header_fields(usergroup)
        return usergroup



# encoding: utf-8

# custom error class for the API


class PackageNotFound(Exception):
    pass


class PlanNotFound(Exception):
    pass


class InsufficientResourceError(Exception):
    pass


class ServiceBindingError(Exception):
    def __init__(self, msg=None):
        Exception.__init__(self, msg)


class APIError(Exception):
    def __init__(self, msg=None):
        Exception.__init__(self, msg)


class ResourceNotFoundError(Exception):
    def __init__(self, msg=None):
        Exception.__init__(self, msg)


class UnauthorizedError(Exception):
    def __init__(self, msg=None):
        Exception.__init__(self, msg)


class ForbiddenError(Exception):
    def __init__(self, msg=None):
        Exception.__init__(self, msg)


class BadRequestError(Exception):
    def __init__(self, msg=None):
        Exception.__init__(self, msg)


class OperationNotAllowedError(Exception):
    def __init__(self, msg=None):
        Exception.__init__(self, msg)


class DeviceNotFoundException(Exception):
    def __init__(self, msg=None):
        Exception.__init__(self, msg)


class ParameterMissingException(Exception):
    def __init__(self, msg=None):
        Exception.__init__(self, msg)


class LabelNotFoundException(Exception):
    def __init__(self, msg=None):
        Exception.__init__(self, msg)


class ConfigNotFoundException(Exception):
    def __init__(self, msg=None):
        Exception.__init__(self, msg)


class DeserializationException(Exception):
    def __init__(self, msg=None):
        Exception.__init__(self, msg)


class DeploymentRunningException(Exception):
    def __init__(self, msg=None):
        Exception.__init__(self, msg)


class CommandException(Exception):
    def __init__(self, msg=None):
        Exception.__init__(self, msg)


class InvalidCommandException(Exception):
    def __init__(self, msg=None):
        Exception.__init__(self, msg)


class InvalidAuthTokenException(Exception):
    def __init__(self, msg=None):
        Exception.__init__(self, msg)


class InvalidParameterException(Exception):
    def __init__(self, msg=None):
        Exception.__init__(self, msg)


class UnknownTopicException(Exception):
    def __init__(self, msg=None):
        Exception.__init__(self, msg)


class UnknownTopicStatusException(Exception):
    def __init__(self, msg=None):
        Exception.__init__(self, msg)


class RetriesExhausted(Exception):
    def __init__(self, msg=None):
        Exception.__init__(self, msg)


class DeploymentNotRunningException(Exception):
    """
    :ivar deployment_status: Deployment status object retrieved from the last poll
    """

    def __init__(self, msg, deployment_status=None):
        self.deployment_status = deployment_status
        Exception.__init__(self, msg)


class ComponentNotFoundException(Exception):
    def __init__(self, msg=None):
        Exception.__init__(self, msg)


class AliasNotProvidedException(Exception):
    def __init__(self, msg=None):
        Exception.__init__(self, msg)


class DuplicateAliasException(Exception):
    def __init__(self, msg=None):
        Exception.__init__(self, msg)


class LogsUUIDNotFoundException(Exception):
    def __init__(self, msg=None):
        Exception.__init__(self, msg)


class InternalServerError(Exception):
    def __init__(self, msg=None):
        Exception.__init__(self, msg)


class ConflictError(Exception):
    def __init__(self, msg=None):
        Exception.__init__(self, msg)


class BuildFailed(Exception):
    def __init__(self, msg=None):
        Exception.__init__(self, msg)


class BuildOperationFailed(Exception):
    def __init__(self, msg):
        Exception.__init__(self, msg)


class InvalidJSONError(Exception):
    def __init__(self, file_path=None):
        msg = "Invalid JSON"
        if file_path:
            msg += ": {}".format(file_path)

        Exception.__init__(self, msg)


class InvalidYAMLError(Exception):
    def __init__(self, file_path=None):
        msg = "Invalid YAML"
        if file_path:
            msg += ": {}".format(file_path)

        Exception.__init__(self, msg)

class UploadError(Exception):
    def __init__(self, msg, file_path):
        full = "Uploading file {} failed".format(file_path)
        if msg:
            full += ": {}".format(msg)

        Exception.__init__(self, full)



# encoding: utf-8
from __future__ import absolute_import
from rapyuta_io.utils.utils import create_auth_header, prepend_bearer_to_auth_token


# Base class for Client connection configuration


class ConnectionConfig(object):
    def __init__(self, auth_token, project):
        self._auth_token = prepend_bearer_to_auth_token(auth_token)
        self._project = project
        self._headers = create_auth_header(self._auth_token,
                                           self._project)

    def set_project(self, project):
        self._project = project
        self._headers = create_auth_header(self._auth_token,
                                           self._project)


class CatalogConfig(ConnectionConfig):
    def __init__(self, catalog_api_host, auth_token, project):
        super(CatalogConfig, self).__init__(auth_token, project)
        self._catalog_api_host = catalog_api_host

    def _get_api_path(self):
        return self._catalog_api_host



from __future__ import absolute_import

import enum

from rapyuta_io.utils import RestClient
from rapyuta_io.utils.rest_client import HttpMethod
from rapyuta_io.utils.utils import get_api_response_data


class AuthTokenLevel(str, enum.Enum):
    def __str__(self):
        return str(self.value)

    LOW = "low"
    MED = "med"
    HIGH = "high"


class RIPClient:
    AUTH_TOKEN_PATH = '/user/login'

    def __init__(self, rip_host):
        self._rip_host = rip_host

    def get_auth_token(self, email, password, token_level=AuthTokenLevel.LOW):
        """
        Fetches an auth token.

        You can fetch a new auth token and also specify the token validity
        with the token_level argument. It takes one of the following values
        `low`, `med` and `high` which correspond to 24h, 7d and 90d of token
        validity respectively.
        """
        url = self._rip_host + self.AUTH_TOKEN_PATH
        payload = {'email': email, 'password': password}
        response = RestClient(url).method(HttpMethod.POST).query_param(
            {'type': token_level}).execute(payload)
        data = get_api_response_data(response, parse_full=True)
        return data['data']['token']



# encoding: utf-8
from __future__ import absolute_import
import json
import re
import six
from time import mktime
from datetime import datetime
from dateutil.parser import parse

from rapyuta_io.utils import ObjDict, InvalidCommandException, InvalidParameterException
from rapyuta_io.utils.object_converter import ObjBase
from rapyuta_io.utils.settings import ENV_VAR_REGEX_PATTERN, SHARED_URL_PATH
from rapyuta_io.utils.utils import is_empty

DEFAULT_LOG_UPLOAD_BANDWIDTH = 1 * 1024 * 1024


class DeviceConfig(ObjDict):
    """
    DeviceConfig class represents configuration of a device. Member variables of the class
    represent the properties of device.

    :ivar id: Id of the configuration.
    :ivar key: Configuration key.
    :ivar value: Value of the configuration key.
    """
    DEFAULT_DEVICE_CONFIG_VARIABLES = ('runtime', 'ros_workspace', 'ros_distro')

    def __init__(self, *args, **kwargs):
        super(ObjDict, self).__init__(*args, **kwargs)

    def is_deletable(self):
        if self.key in self.DEFAULT_DEVICE_CONFIG_VARIABLES:
            return False

        if self.key in ['runtime_docker', 'runtime_preinstalled']:
            return False

        return True

    def is_updatable(self):
        if self.key == self.DEFAULT_DEVICE_CONFIG_VARIABLES[0]:
            return False
        return True


class Command(ObjDict):
    """
    Command class represent command that is executed on a device.

    :ivar shell: Represents the shell where it is going to execute
    :ivar env: List of environment variables.
    :ivar bg: Boolean value specifying whether the execution runs on the background or not
    :ivar run_async: Boolean value to specify to run command in async mode i.e., to run command and return its jid (not depending on completion of the command).
    :ivar runas: Run the command as a specific user
    :ivar cmd: Command to execute on the device
    :ivar pwd: Present working directory
    :ivar cwd: Current working directory

    :Note: **pwd** is deprecated. Although backward compatibility is supported, it is recommended to use **cwd** instead.

    """

    def __init__(self, cmd, shell=None, env=None, bg=False, run_async=False, runas=None, pwd=None, cwd=None, timeout=300):
        super(ObjDict, self).__init__()
        if env is None:
            env = dict()
        self.cmd = cmd
        self.shell = shell
        self.env = env
        self.bg = bg
        self.run_async = run_async
        self.runas = runas
        self.cwd = pwd
        if cwd is not None:
            self.cwd = cwd
        self.timeout = timeout
        self.validate()

    def validate(self):
        if is_empty(self.cmd) or not isinstance(self.cmd, six.string_types):
            raise InvalidCommandException("Invalid execution command")
        if self.shell is not None and not isinstance(self.shell, six.string_types):
            raise InvalidCommandException("Invalid shell")
        if self.bg is not None and not isinstance(self.bg, bool):
            raise InvalidCommandException("Invalid background option")
        if self.run_async is not None and not isinstance(self.run_async, bool):
            raise InvalidCommandException("Invalid run asynchronous option")
        if self.runas is not None and not isinstance(self.runas, six.string_types):
            raise InvalidCommandException("Invalid runas option")
        if self.cwd is not None and not isinstance(self.cwd, six.string_types):
            raise InvalidCommandException("Invalid cwd option")
        if self.env:
            if isinstance(self.env, dict):
                for env_var in self.env:
                    if not re.match(ENV_VAR_REGEX_PATTERN, env_var):
                        raise InvalidCommandException('Invalid environment variables')
                return
            raise InvalidCommandException('Invalid environment variables')
        if  self.timeout <= 0:
            raise InvalidCommandException("Invalid timeout value")

    def to_json(self):
        # TODO: we need to rewrite this function.
        js = json.loads(json.dumps(self))
        ret = {}
        for key, value in js.items():
            if value is not None and str(value) != '':
                ret[key] = value
        return ret


class TopicsStatus(ObjDict):
    """
    Topic class represents the status - subscribed and unsubscribed - for logs and metrics

    :ivar master_up: Boolean represents whether the master is running or not
    :ivar subscribed: List of subscribed topics
    :ivar unsubscribed: List of unsubscribed topics

     """

    def __init__(self, *args, **kwargs):
        super(ObjDict, self).__init__(*args, **kwargs)


class Label(ObjDict):
    """ Label class represents labels associated with a device.

         :ivar id: Integer represents the id of the label
         :ivar key: Key or label name
         :ivar value: Value of the label

         """

    def __init__(self, *args, **kwargs):
        super(ObjDict, self).__init__(*args, **kwargs)


class Metric(ObjDict):
    """
    Class represents current status of subscription of the metric

    :ivar name: Name of the metric
    :ivar kind: metric
    :ivar config: Metric collection configuration
    """

    def __init__(self, *args, **kwargs):
        super(ObjDict, self).__init__(*args, **kwargs)


class LogsUploadRequest(ObjDict):
    """
     Request class for log upload API

     :ivar device_path: absolute path of file on device
     :ivar file_name: Name of the file in cloud storage. If not provided, file name is derived from the device_path
     :ivar max_upload_rate: network bandwidth to be used for upload
     :ivar override: If true overrides the destination file
     :ivar purge_after: If true purges the log file after upload
     :ivar metadata: Key/value to be associated with log file
    """

    def __init__(self, device_path, file_name='', max_upload_rate=DEFAULT_LOG_UPLOAD_BANDWIDTH, override=False, purge_after=False,
                 metadata=None):
        super(ObjDict, self).__init__()
        self.device_path = device_path
        self.file_name = file_name
        self.max_upload_rate = max_upload_rate
        self.override = override
        self.purge_after = purge_after
        self.metadata = metadata if metadata else {}
        self._validate()

    def _validate(self):
        if is_empty(self.device_path) or not isinstance(self.device_path, six.string_types):
            raise InvalidParameterException('file must be valid')

    def to_json(self):
        js = self.to_dict()
        ret = {}
        for key, value in js.items():
            if value is not None and str(value) != '':
                ret[key] = value
        return ret


class LogUploadStatus(ObjDict):
    """
    This class instance represent current status of the upload request

    :ivar status: Represents the current file upload status
    :ivar error_message: Any error message during file upload
    :ivar creator: User guid who initiated the upload operation
    :ivar created_at: Timestamp at which logs upload initiated
    :ivar request_uuid: An uniquely identifier associated with current upload
    :ivar filename: Name of the file which is to be uploaded
    :ivar updated_at: Timestamp at which file is updated
    """

    def __init__(self, *args, **kwargs):
        super(ObjDict, self).__init__(*args, **kwargs)
        # The SharedURL class is based on the ObjBase but LogUploadStatus is not. This logic acts as the glue that
        # deserializes the shared_urls and puts it back in the ObjDict.
        if 'shared_urls' in self:
            url_list = []
            for shared_url in self['shared_urls']:
                url_list.append(SharedURL.deserialize(shared_url))
            self['shared_urls'] = url_list


class LogUploads(ObjDict):
    """
    This class represents the status and other attributes associated with upload files.

    :ivar status: Represents the current upload status
    :ivar content_length: Length in bytes
    :ivar creator: User guid who initiated the upload operation
    :ivar created_at: Timestamp at which logs upload initiated
    :ivar request_uuid: UUID associated with the request
    :ivar creation_time: Timestamp at which file upload is completed
    :ivar updated_at: Timestamp at which last file updated happened
    :ivar filename: Name of the file which is to be uploaded
    :ivar last_modified: represents the time at which last modification done to the file on cloud
    :ivar metadata: key/value pair associated with upload request
    """

    def __init__(self, *args, **kwargs):
        super(ObjDict, self).__init__(*args, **kwargs)


class SharedURL(ObjBase):
    """
    This class represents the Shared URL for Logs uploaded from the Device.

    :param request_uuid: Request GUID of the Log file
    :type request_uuid: str
    :param expiry_time: Datetime object for the time at which Shared URL is supposed to expire.
    :type expiry_time: :py:class:`datetime.datetime`
    :param creator: User GUID that created the Shared URL
    :type creator: str
    :param created_at: Timestamp at which Shared URL was created
    :type created_at: str
    """

    def __init__(self, request_uuid, expiry_time):
        self.validate(request_uuid, expiry_time)
        self.request_uuid = request_uuid
        self.expiry_time = expiry_time
        self.url_uuid = None
        self.creator = None
        self.created_at = None

    @property
    def url(self):
        """
        The generated public URL.

        :raises: :py:class:`~rapyuta_io.utils.errors.InvalidParameterException`: If the Shared URL is not created
        """
        if not hasattr(self, '_device_api_host'):
            raise InvalidParameterException('SharedURL must be created first')
        return '{}/{}/{}'.format(self._device_api_host, SHARED_URL_PATH, self.url_uuid)

    def serialize(self):
        return {
            'expiry_time': mktime(self.expiry_time.timetuple())
        }

    def get_serialize_map(self):
        pass

    def get_deserialize_map(self):
        return {
            'expiry_time': ('expiry_time', lambda data: parse(data) if data is not None else None),
            'created_at': 'created_at',
            'creator': 'creator',
            'url_uuid': 'url_uuid'
        }

    @staticmethod
    def validate(request_uuid, expiry_time):
        if not (isinstance(request_uuid, six.binary_type) or isinstance(request_uuid, six.string_types)) or request_uuid == '':
            raise InvalidParameterException('request_uuid must be a non-empty string')
        if not isinstance(expiry_time, datetime) or expiry_time is None:
            raise InvalidParameterException('expiry_time must be a datetime.datetime')



# encoding: utf-8
from __future__ import absolute_import

import enum
from platform import python_implementation, python_version
from time import sleep

import requests
from requests.exceptions import RequestException

import rapyuta_io
from rapyuta_io.utils import APIError

DEFAULT_RETRY_COUNT = 4
WAIT_TIME_IN_SEC = 1


requests.utils.default_user_agent = lambda: 'rapyuta_io/{} {}/{} python-requests/{}'.format(
    rapyuta_io.__version__, python_implementation(), python_version(), requests.__version__)


class HttpMethod(str, enum.Enum):

    def __str__(self):
        return str(self.value)

    GET = 'GET'
    POST = 'POST'
    PUT = 'PUT'
    PATCH = 'PATCH'
    DELETE = 'DELETE'
    OPTIONS = 'OPTIONS'


class RestClient:
    def __init__(self, url):
        self._url = url
        self._retry_limit = DEFAULT_RETRY_COUNT
        self._retry_count = 0
        self._method = HttpMethod.GET.value
        self._headers = {}
        self._query_params = {}

    def url(self, url):
        self._url = url
        return self

    def method(self, http_method):
        self._method = http_method.value
        return self

    def headers(self, headers):
        self._headers = headers
        return self

    def retry(self, retry_limit):
        if type(retry_limit) is int:
            self._retry_limit = retry_limit
        return self

    def query_param(self, query_param):
        self._query_params = query_param
        return self

    def _request(self, payload, raw=False):
        kwargs = {'method': self._method, 'url': self._url,
                  'headers': self._headers, 'params': self._query_params,
                  'timeout': (30, 150) } # Configures ConnectTimeout(30sec) and ReadTimeout(150sec)
        if raw:
            response = requests.request(data=payload, **kwargs)
            return response
        response = requests.request(json=payload, **kwargs)
        return response

    def execute(self, payload=None, raw=False):
        while True:
            try:
                # It will not be respecting the Retry Limit, when the server
                # returns Internal Server Error (500) for a GET Request. This is
                # because in most places of the SDK we are setting Retry Limit
                # explicitly to 0. The plan is to deprecate the Retry Limit in
                # future.
                response = self._request(payload, raw)
                if self._method != HttpMethod.GET.value or \
                        response.status_code != requests.codes.INTERNAL_SERVER_ERROR or \
                        self._retry_count >= DEFAULT_RETRY_COUNT:
                    return response
            except RequestException as err:
                if self._retry_count >= self._retry_limit:
                    raise APIError("Error occurred for URL {}: {}".format(self._url, err))

            sleep(WAIT_TIME_IN_SEC)
            self._retry_count += 1



from __future__ import absolute_import
from datetime import datetime, timedelta

import enum
import six
import pytz

from rapyuta_io.utils.error import InvalidParameterException
from rapyuta_io.utils.object_converter import ObjBase, enum_field, list_field
from rapyuta_io.utils.utils import valid_list_elements


def _use_clickhouse(start_datetime):
    use_clickhouse_window = timedelta(days=7)
    if start_datetime.tzinfo:
        interval = datetime.now(tz=pytz.UTC) - start_datetime.astimezone(tz=pytz.UTC)
    else:
        interval = datetime.now() - start_datetime
    return interval > use_clickhouse_window


class _StepIntervalInflux(str, enum.Enum):

    def __str__(self):
        return str(self.value)

    ONE_SECOND = '1s'
    TEN_SECONDS = '10s'
    THIRTY_SECONDS = '30s'
    ONE_MINUTE = '1m'
    FIVE_MINUTES = '5m'
    FIFTEEN_MINUTES = '15m'


class _StepIntervalClickhouse(str, enum.Enum):

    def __str__(self):
        return str(self.value)

    ONE_MINUTE = '1m'
    FIVE_MINUTES = '5m'
    FIFTEEN_MINUTES = '15m'


class StepInterval(str, enum.Enum):
    """
    StepInterval may be one of: \n
    StepInterval.ONE_SECOND (1s) \n
    StepInterval.TEN_SECONDS (10s)  \n
    StepInterval.THIRTY_SECONDS (30s) \n
    StepInterval.ONE_MINUTE (1m) \n
    StepInterval.FIVE_MINUTES (5m) \n
    StepInterval.FIFTEEN_MINUTES (15m) \n
    """

    def __str__(self):
        return str(self.value)

    ONE_SECOND = '1s'
    TEN_SECONDS = '10s'
    THIRTY_SECONDS = '30s'
    ONE_MINUTE = '1m'
    FIVE_MINUTES = '5m'
    FIFTEEN_MINUTES = '15m'


class MetricFunction(str, enum.Enum):
    """
    MetricFunction may be one of: \n
    MetricFunction.COUNT (count)\n
    MetricFunction.DISTINCT (distinct) \n
    MetricFunction.MEAN (mean) \n
    MetricFunction.MEDIAN (median) \n
    MetricFunction.MODE (mode) \n
    MetricFunction.AVG (avg) \n

    MetricFunction.STDDEV (stddev) \n
    MetricFunction.DERIVATIVE (derivative) \n

    MetricFunction.MAX (max) \n
    MetricFunction.MIN (min) \n
    MetricFunction.FIRST (first) \n
    MetricFunction.LAST (last) \n
    MetricFunction.PERCENTILE_95 (percentile_95) \n
    MetricFunction.PERCENTILE_99 (percentile_99) \n

    """

    def __str__(self):
        return str(self.value)

    COUNT = 'count'
    DISTINCT = 'distinct'
    MEAN = 'mean'
    MEDIAN = 'median'
    MODE = 'mode'
    AVG = 'avg'

    # Transformations
    STDDEV = 'stddev'
    DERIVATIVE = 'derivative'

    # Selectors
    MAX = 'max'
    MIN = 'min'
    FIRST = 'first'
    LAST = 'last'
    PERCENTILE_95 = 'percentile_95'
    PERCENTILE_99 = 'percentile_99'


class SortOrder(str, enum.Enum):
    """
    SortOrder may be one of: \n
    SortOrder.ASC (asc) \n
    SortOrder.DESC (desc) \n
    """

    def __str__(self):
        return str(self.value)

    ASC = 'asc'
    DESC = 'desc'


class Entity(str, enum.Enum):
    """
    Entity may be one of: \n
    Entity.DEVICE (device) \n
    Entity.DEPLOYMENT (deployment) \n
    Entity.PROJECT (project) \n
    Entity.ORGANIZATION (organization) \n
    """

    def __str__(self):
        return str(self.value)

    DEVICE = 'device'
    DEPLOYMENT = 'deployment'
    PROJECT = 'project'
    ORGANIZATION = 'organization'


class MetricOperation(ObjBase):
    """
    MetricsOperation class that defines a function over a metric

    :ivar function: function to applied on metric
    :vartype function: :py:class:`~rapyuta_io.clients.metrics.MetricFunction`
    :ivar metric_name: name of the metric
    :vartype metric_name: str

    :param function: function to applied on metric
    :type function: :py:class:`~rapyuta_io.clients.metrics.MetricFunction`
    :param metric_name: name of the metric
    :type metric_name: str
    """
    def __init__(self, function, metric_name):
        self.validate(function, metric_name)
        self.function = function
        self.metric_name = metric_name

    @staticmethod
    def validate(function, metric_name):
        if function not in list(MetricFunction.__members__.values()):
            raise InvalidParameterException('function should be member of MetricFunction')
        if not isinstance(metric_name, six.string_types):
            raise InvalidParameterException('metric name should be a string')

    def get_deserialize_map(self):
        return {}

    def get_serialize_map(self):
        return {
            'function': 'function',
            'metric_name': 'metric_name'
        }


class QueryMetricsRequest(ObjBase):
    """
    QueryMetricsRequest class

    :ivar from_datetime: start time of querying metrics
    :vartype from_datetime: :py:class:`~datetime.datetime`
    :ivar to_datetime: end time of querying metrics
    :vartype to_datetime: :py:class:`~datetime.datetime`
    :ivar step_interval: time interval to group data
    :vartype step_interval: :py:class:`~rapyuta_io.clients.metrics.StepInterval`
    :ivar ~.metrics: list of metrics
    :vartype ~.metrics: list(:py:class:`~rapyuta_io.clients.metrics.MetricOperation`)
    :ivar tags: key pair of tags (must include tenant_id and organization_id)
    :vartype tags: dict
    :ivar sort: ordering to sort the metrics
    :vartype sort: :py:class:`~rapyuta_io.clients.metrics.SortOrder`
    :ivar groupby: list of tags to group data
    :vartype groupby: list(str)

    :param from_datetime: start time of querying metrics
    :type from_datetime: :py:class:`~datetime.datetime`
    :param to_datetime: end time of querying metrics
    :type to_datetime: :py:class:`~datetime.datetime`
    :param step_interval: time interval to group data
    :type step_interval: :py:class:`~rapyuta_io.clients.metrics.StepInterval`
    :param metrics: list of metrics
    :type metrics: list(:py:class:`~rapyuta_io.clients.metrics.MetricOperation`)
    :param tags: key pair of tags (must include tenant_id and organization_id)
    :type tags: dict
    :param sort: ordering to sort the metrics
    :type sort: :py:class:`~rapyuta_io.clients.metrics.SortOrder`
    :param groupby: list of tags to group data
    :type groupby: list(str)
    """

    TENANT_ID_TAG = 'tenant_id'
    ORGANIZATION_ID_TAG = 'organization_id'

    def __init__(self, from_datetime, to_datetime, step_interval, metrics, tags=None, sort=None, groupby=None):
        self.validate(from_datetime, to_datetime, step_interval, metrics, tags, sort, groupby)
        self.from_datetime = from_datetime
        self.to_datetime = to_datetime
        self.step_interval = step_interval
        self.metrics = metrics
        self.sort = sort or SortOrder.ASC
        self.groupby = groupby
        self.tags = tags or {}

    def validate(self, from_datetime, to_datetime, step_interval, metrics, tags, sort, groupby):

        if not isinstance(from_datetime, datetime):
            raise InvalidParameterException('from_datetime should be a datetime object')
        if not isinstance(to_datetime, datetime):
            raise InvalidParameterException('to_datetime should be a datetime object')
        if not valid_list_elements(metrics, MetricOperation) or len(metrics) < 1:
            raise InvalidParameterException('metrics should be non-empty list of MetricOperation objects')
        if sort is not None and sort not in list(SortOrder.__members__.values()):
            raise InvalidParameterException('sort should be a member of SortOrder enum')
        if groupby is not None and not valid_list_elements(groupby, six.string_types):
            raise InvalidParameterException('grouby should be list of strings')
        if tags is not None and not isinstance(tags, dict):
            raise InvalidParameterException('tags should be key value of tags and values')
        if step_interval not in list(StepInterval.__members__.values()):
            raise InvalidParameterException('step_interval should be a member of StepInterval enum')

        step_interval_enums = _StepIntervalInflux
        if _use_clickhouse(from_datetime):
            step_interval_enums = _StepIntervalClickhouse

        choices = [v.value for v in step_interval_enums]
        if step_interval.value not in choices:
            raise InvalidParameterException('step_interval must be one of 1m, 5m, 15m '
                                            'for from_datetime before 7 days from today')

        if from_datetime > to_datetime:
            raise InvalidParameterException('from_datetime must be less than to_datetime')

    def get_deserialize_map(self):
        return {}

    def serialize(self):
        serialized = super(QueryMetricsRequest, self).serialize()
        serialized.update({'from': self.from_datetime.isoformat(),
                           'to': self.to_datetime.isoformat()})
        return serialized

    def get_serialize_map(self):
        return {
            'step': 'step_interval',
            'metrics': 'metrics',
            'sort': 'sort',
            'groupby': 'groupby',
            'tags': 'tags',
        }


class QueryMetricsResponse(ObjBase):
    """
    QueryMetricsResponse class

    :ivar columns: list of columns object
    :vartype columns: :py:class:`~rapyuta_io.clients.metrics.Column`
    :ivar rows: list of metrics values
    :vartype rows: :py:class:`list(list(float))`
    """
    def __init__(self):
        self.rows = []
        self.columns = []

    def get_serialize_map(self):
        return {}

    def get_deserialize_map(self):
        return {
            'rows': 'rows',
            'columns': list_field('columns', Column),
        }

    def to_row_column_format(self):
        """
        Returns rows and columns in a format that can be passed to `pandas.DataFrame()`.

        :rtype: tuple(generator, list)
        """
        return zip(*self.rows), list(map(str, self.columns))


class Column(ObjBase):
    """
    Column class

    :ivar name: name of metric
    :vartype name: str
    :ivar function: applied metric function
    :vartype function: :py:class:`~rapyuta_io.clients.metrics.MetricFunction`
    :ivar metric_group: metric group
    :vartype metric_group: str
    :ivar tag_names: list of tag names
    :vartype tag_names: str
    :ivar tag_values: list of tag values
    :vartype tag_values: str
    """
    def __init__(self):
        self.name = None
        self.function = None
        self.metric_group = None
        self.tag_names = None
        self.tag_values = None

    def get_serialize_map(self):
        return {}

    def get_deserialize_map(self):
        return {
            'name': 'name',
            'function': enum_field('function', MetricFunction),
            'metric_group': 'metric_group',
            'tag_names': 'tag_names',
            'tag_values': 'tag_values',
        }

    def __str__(self):
        if self.name == 'timestamp':
            return self.name

        if self.tag_values:
            tag_pair = ['{}={}'.format(k, v) for k, v in zip(self.tag_names, self.tag_values)]
            tags = '{' + ','.join(tag_pair) + '}'
            return '{}({}.{}){}'.format(self.function, self.metric_group, self.name, tags)

        return '{}({}.{})'.format(self.function, self.metric_group, self.name)


class _MetricsMetaRequest(object):
    """
    Metrics Metadata Query

    :ivar entity: one of Entity enum
    :vartype entity: :py:class:`~rapyuta_io.clients.metrics.Entity`
    :ivar entity_id: value of entity
    :vartype entity_id: str
    :ivar start_date: filter from start_date
    :vartype start_date: :py:class:`~datetime.datetime`
    :ivar end_date: filter to end_date
    :vartype end_date: :py:class:`~datetime.datetime`

    :param entity: one of Entity enum
    :type entity: :py:class:`~rapyuta_io.clients.metrics.Entity`
    :param entity_id: value of entity
    :type entity_id: str
    :param start_date: filter from start_date
    :type start_date: :py:class:`~datetime.datetime`
    :param end_date: filter to end_date
    :type end_date: :py:class:`~datetime.datetime`
    """

    def __init__(self, entity, entity_id, start_date, end_date):
        self.validate(entity, entity_id, start_date, end_date)
        self.entity = entity
        self.entity_id = entity_id
        self.start_date = start_date
        self.end_date = end_date

    @staticmethod
    def validate(entity, entity_id, start_date, end_date):
        if entity not in list(Entity.__members__.values()):
            raise InvalidParameterException('entity should be a member of Entity enum')
        if not isinstance(entity_id, six.string_types) or not entity_id:
            raise InvalidParameterException('entity_id non empty string')
        if not isinstance(start_date, datetime):
            raise InvalidParameterException('start_date should be a datetime object')
        if not isinstance(end_date, datetime):
            raise InvalidParameterException('end_date should be a datetime object')

        if start_date > end_date:
            raise InvalidParameterException('start_date must be less than end_date')

        if entity == Entity.ORGANIZATION and not _use_clickhouse(start_date):
            raise InvalidParameterException('Entity.ORGANIZATION cannot be queried '
                                            'for start_date before 7 days from today')

    def to_dict(self):
        return {
            'entity': self.entity,
            'entity_id': self.entity_id,
            'start_date': self.start_date.isoformat(),
            'end_date': self.end_date.isoformat(),
        }


class Metric(ObjBase):
    """
    Metric class, has metric_group and list of metric names

    :ivar metric_group: name of the metric
    :vartype metric_group: str
    :ivar metric_names: list of metrics
    :vartype metric_names: list(str)
    """

    def get_serialize_map(self):
        return {}

    def get_deserialize_map(self):
        return {
            'metric_group': 'metric_group',
            'metric_names': 'metric_names'
        }


class Tags(ObjBase):
    """
    Tags class, has metric_group and list of tags

    :ivar metric_group: name of the metric
    :vartype metric_group: str
    :ivar tags: list of tags
    :vartype tags: list(str)
    """

    def get_serialize_map(self):
        return {}

    def get_deserialize_map(self):
        return {
            'metric_group': 'metric_group',
            'tags': 'tags'
        }


class ListMetricsRequest(_MetricsMetaRequest):

    __doc__ = _MetricsMetaRequest.__doc__.replace('Metrics Metadata Query', 'List Metrics Request')

    def __init__(self, entity, entity_id, start_date, end_date):
        super(ListMetricsRequest, self).__init__(entity, entity_id, start_date, end_date)


class ListTagKeysRequest(_MetricsMetaRequest):

    __doc__ = _MetricsMetaRequest.__doc__.replace('Metrics Metadata Query', 'List Tag Keys Request')

    def __init__(self, entity, entity_id, start_date, end_date):
        super(ListTagKeysRequest, self).__init__(entity, entity_id, start_date, end_date)


class ListTagValuesRequest(_MetricsMetaRequest):
    """
    List Tag Values Request

    :ivar entity: one of Entity enum
    :vartype entity: :py:class:`~rapyuta_io.clients.metrics.Entity`
    :ivar entity_id: value of entity
    :vartype entity_id: str
    :ivar tag: value of the tag key
    :vartype tag: str
    :ivar start_date: filter from start_date
    :vartype start_date: :py:class:`~datetime.datetime`
    :ivar end_date: filter to end_date
    :vartype end_date: :py:class:`~datetime.datetime`

    :param entity: one of Entity enum
    :type entity: :py:class:`~rapyuta_io.clients.metrics.Entity`
    :param entity_id: value of entity
    :type entity_id: str
    :param tag: value of the tag key
    :type tag: str
    :param start_date: filter from start_date
    :type start_date: :py:class:`~datetime.datetime`
    :param end_date: filter to end_date
    :type end_date: :py:class:`~datetime.datetime`
    """

    def __init__(self, entity, entity_id, tag, start_date, end_date):
        self.validate_tag(tag)
        super(ListTagValuesRequest, self).__init__(entity, entity_id, start_date, end_date)
        self.tag = tag

    @staticmethod
    def validate_tag(tag):
        if not isinstance(tag, six.string_types) or not tag:
            raise InvalidParameterException('tag non empty string')



from __future__ import absolute_import
from abc import ABCMeta, abstractmethod
import enum
import six


class ObjBase(six.with_metaclass(ABCMeta)):
    @abstractmethod
    def get_serialize_map(self):
        """
        Returns Dict[api_field, attr_name]

        :return: dict
        """
    @abstractmethod
    def get_deserialize_map(self):
        """
        Returns Dict[attr_name, api_field or (api_field, callable)]

        Deserializes response into the object based on the Map. For arbitrary logic, a tuple of (api_field, callable)
        can be passed. Callable will be called with api_field_value if present.

        :return: dict
        """

    def serialize(self):
        res = {}
        serialization_map = self.get_serialize_map()
        for api_field, attr_name in serialization_map.items():
            attr_val = getattr(self, attr_name)
            if attr_val is None:
                continue
            if isinstance(attr_val, ObjBase):
                res[api_field] = attr_val.serialize()
            elif isinstance(attr_val, list) and all([isinstance(x, ObjBase) for x in attr_val]):
                res[api_field] = [v.serialize() for v in attr_val]
            elif isinstance(attr_val, enum.Enum):
                res[api_field] = attr_val.value
            else:
                res[api_field] = attr_val
        return res

    @classmethod
    def deserialize(cls, data, obj=None, only=None):
        if not obj:
            obj = cls.__new__(cls)
        if only and (not isinstance(only, list) or any([x for x in only if not isinstance(x, str)])):
            raise Exception('only should be a list of string')
        deserialization_map = obj.get_deserialize_map()
        for attr_name, val in deserialization_map.items():
            if only and attr_name not in only:
                continue
            if isinstance(val, tuple):
                api_field, callable_ = val
                deserialized_val = callable_(data.get(api_field))
                setattr(obj, attr_name, deserialized_val)
            else:
                setattr(obj, attr_name, data.get(val))
        return obj


def nested_field(api_field, cls):
    if not issubclass(cls, ObjBase):
        raise Exception('{} should be a subclass of ObjBase'.format(cls.__name__))
    return api_field, lambda data: cls.deserialize(data) if data is not None else None


def list_field(api_field, cls):
    if not issubclass(cls, ObjBase):
        raise Exception('{} should be a subclass of ObjBase'.format(cls.__name__))

    def deserialize(data):
        if not isinstance(data, list):
            return data
        return [cls.deserialize(d) for d in data]

    return api_field, deserialize


def enum_field(api_field, cls):
    if not issubclass(cls, enum.Enum):
        raise Exception('{} should be a subclass of Enum'.format(cls.__name__))

    def deserialize(data):
        if data not in list(cls.__members__.values()):
            return data
        return cls(data)

    return api_field, deserialize



from __future__ import absolute_import

import enum
import errno
import hashlib
import json
import mimetypes
import os
import tempfile
from concurrent import futures
from os import listdir, makedirs
from os.path import isdir, join
from shutil import rmtree, copyfile

import six

from rapyuta_io.utils import RestClient, InvalidParameterException, ConfigNotFoundException
from rapyuta_io.utils.error import InvalidJSONError, InvalidYAMLError, UploadError
from rapyuta_io.utils.rest_client import HttpMethod
from rapyuta_io.utils.settings import PARAMSERVER_API_TREE_PATH, PARAMSERVER_API_TREEBLOBS_PATH, \
    PARAMSERVER_API_FILENODE_PATH
from rapyuta_io.utils.utils import (
    create_auth_header,
    prepend_bearer_to_auth_token,
    get_api_response_data,
    validate_list_of_strings,
    parse_json,
    parse_yaml
)


class _Node(str, enum.Enum):

    def __str__(self):
        return str(self.value)

    File = 'FileNode'
    Value = 'ValueNode'
    Attribute = 'AttributeNode'
    Folder = 'FolderNode'


class _ParamserverClient:
    """
    Internal client for paramserver. Not for public use.
    """
    yaml_content_type = 'text/yaml'
    json_content_type = 'application/json'
    default_binary_content_type = "application/octet-stream"
    max_non_binary_size = 128 * 1024

    def __init__(self, auth_token, project, core_api_host):
        self._auth_token = auth_token
        self._headers = create_auth_header(prepend_bearer_to_auth_token(auth_token), project)
        self._core_api_host = core_api_host

    def set_project(self, project_guid):
        self._headers = create_auth_header(prepend_bearer_to_auth_token(self._auth_token), project_guid)

    @staticmethod
    def get_md5_checksum(data):
        md5_hash = hashlib.md5()
        md5_hash.update(data)
        return md5_hash.hexdigest()

    def create_file(self, tree_path, filedata, retry_limit=0, content_type=None):
        content_type = content_type if content_type else self.yaml_content_type
        url = self._core_api_host + PARAMSERVER_API_TREE_PATH + tree_path
        payload = {'type': _Node.File, 'data': filedata, 'contentType': content_type}
        response = RestClient(url).method(HttpMethod.PUT).headers(self._headers).retry(retry_limit).execute(payload)
        return get_api_response_data(response, parse_full=True)

    def create_binary_file(self, tree_path, file_path, retry_limit=0):
        url = self._core_api_host + PARAMSERVER_API_FILENODE_PATH + tree_path
        content_type = self.default_binary_content_type
        headers = self._headers.copy()
        guessed_content_type = mimetypes.MimeTypes().guess_type(file_path)
        if len(guessed_content_type) != 0:
            if guessed_content_type[0]:
                content_type = guessed_content_type[0]
            if guessed_content_type[1]:
                headers['Content-Encoding'] = guessed_content_type[1]

        # Override Content-Type for JSON and YAML to allow creating Binary files.
        # This is required for large YAML/JSON files.
        if content_type in (self.json_content_type, self.yaml_content_type):
            content_type = self.default_binary_content_type

        headers.update({'X-Rapyuta-Params-Version': "0",
                        'Content-Type': content_type})

        with open(file_path, 'rb') as f:
            headers.update({'Checksum': self.get_md5_checksum(f.read())})
            f.seek(0)
            response = RestClient(url).method(HttpMethod.PUT).headers(headers).execute(payload=f, raw=True)

        try:
            return get_api_response_data(response, parse_full=True)
        except Exception as e:
            raise UploadError(file_path=file_path, msg=e)

    def create_value(self, tree_path, retry_limit=0):
        url = self._core_api_host + PARAMSERVER_API_TREE_PATH + tree_path
        payload = {'type': _Node.Value}
        response = RestClient(url).method(HttpMethod.PUT).headers(self._headers).retry(retry_limit).execute(payload)
        return get_api_response_data(response, parse_full=True)

    def create_folder(self, tree_path, retry_limit=0):
        url = self._core_api_host + PARAMSERVER_API_TREE_PATH + tree_path
        payload = {'type': _Node.Folder}
        response = RestClient(url).method(HttpMethod.PUT).headers(self._headers).retry(retry_limit).execute(payload)
        return get_api_response_data(response, parse_full=True)

    def create_attribute(self, tree_path, retry_limit=0):
        url = self._core_api_host + PARAMSERVER_API_TREE_PATH + tree_path
        payload = {'type': _Node.Attribute}
        response = RestClient(url).method(HttpMethod.PUT).headers(self._headers).retry(retry_limit).execute(payload)
        return get_api_response_data(response, parse_full=True)

    def create_tree(self, tree_name, delete_existing, retry_limit=0):
        if delete_existing:
            url = self._core_api_host + PARAMSERVER_API_TREE_PATH + tree_name
            response = RestClient(url).method(HttpMethod.DELETE).headers(self._headers).retry(retry_limit).execute()
            get_api_response_data(response, parse_full=True)  # validate 200 response
        return self.create_value(tree_name)

    def process_root_dir(self, executor, rootdir, tree_names, delete_existing_trees):
        listdir_names = listdir(rootdir)
        if tree_names:
            listdir_names = [name for name in listdir_names if name in tree_names]
        dir_futures = {}
        for name in listdir_names:
            if isdir(join(rootdir, name)):
                future = executor.submit(self.create_tree, name, delete_existing_trees)
                dir_futures[future] = (name, 1)
        for future in futures.as_completed(dir_futures):
            exc = future.exception()
            if exc is not None:
                exc.tree_path = dir_futures[future][0]
                raise exc
        return dir_futures

    def process_dir(self, executor, rootdir, tree_path, level, dir_futures, file_futures):
        in_attribute_dir = level % 2 == 0
        for name in listdir(join(rootdir, tree_path)):
            full_path = join(rootdir, tree_path, name)
            new_tree_path = join(tree_path, name)
            if isdir(full_path):
                func = self.create_value if in_attribute_dir else self.create_attribute
                future = executor.submit(func, new_tree_path)
                dir_futures[future] = (new_tree_path, level + 1)
            elif not in_attribute_dir:  # ignore files in attribute directories
                file_stat = os.stat(full_path)
                file_name = os.path.basename(full_path)
                if file_stat.st_size > self.max_non_binary_size:
                    future = executor.submit(self.create_binary_file, new_tree_path, full_path)
                if file_name.endswith('.yaml'):
                    data = parse_yaml(full_path)
                    if self.should_upload_as_binary(data, self.yaml_content_type):
                        future = executor.submit(self.create_binary_file, new_tree_path, full_path)
                    else:
                        future = executor.submit(self.create_file, new_tree_path, data)
                elif file_name.endswith('.json'):
                    data = parse_json(full_path)
                    if self.should_upload_as_binary(data, self.json_content_type):
                        future = executor.submit(self.create_binary_file, new_tree_path, full_path)
                    else:
                        future = executor.submit(self.create_file, new_tree_path, data, content_type=self.json_content_type)
                else:
                    future = executor.submit(self.create_binary_file, new_tree_path, full_path)
                file_futures[future] = new_tree_path
        return dir_futures, file_futures

    def process_folder(self, executor, rootdir, tree_path, level, dir_futures, file_futures):
        for name in listdir(join(rootdir, tree_path)):
            full_path = join(rootdir, tree_path, name)
            new_tree_path = join(tree_path, name)
            if isdir(full_path):
                future = executor.submit(self.create_folder, new_tree_path)
                dir_futures[future] = (new_tree_path, level + 1)
                continue
            file_stat = os.stat(full_path)
            file_name = os.path.basename(full_path)
            if file_stat.st_size > self.max_non_binary_size:
                future = executor.submit(self.create_binary_file, new_tree_path, full_path)
            elif file_name.endswith('.yaml'):
                data = parse_yaml(full_path)
                if self.should_upload_as_binary(data, self.yaml_content_type):
                    future = executor.submit(self.create_binary_file, new_tree_path, full_path)
                else:
                    future = executor.submit(self.create_file, new_tree_path, data)
            elif file_name.endswith('.json'):
                data = parse_json(full_path)
                if self.should_upload_as_binary(data, self.json_content_type):
                    future = executor.submit(self.create_binary_file, new_tree_path, full_path)
                else:
                    future = executor.submit(self.create_file, new_tree_path, data, content_type=self.json_content_type)
            else:
                future = executor.submit(self.create_binary_file, new_tree_path, full_path)
            file_futures[future] = new_tree_path

        return dir_futures, file_futures

    def should_upload_as_binary(self, filedata, content_type):
        """Determines if the file should be uploaded as binary based on the file size

        While the file data may be less than the supported limit, the combined size of
        the API payload is what is stored by paramserver. This method calculates the
        size of the payload and determines if it exceeds the supported limit. If it does,
        the file is uploaded as a binary to the blob store.

        We cannot entirely rely on the file stat result since the file data is sent as
        a string in the API payload and the final size inflates when json.dumps is
        performed on the final payload.
        """
        metadata_size_buffer = 200  # In bytes
        payload = {'type': _Node.File, 'data': filedata, 'contentType': content_type}
        return len(json.dumps(payload)) + metadata_size_buffer > self.max_non_binary_size

    def upload_configurations(self, rootdir, tree_names, delete_existing_trees, as_folder=False):
        self.validate_args(rootdir, tree_names, delete_existing_trees, as_folder)
        with futures.ThreadPoolExecutor(max_workers=15) as executor:
            dir_futures = self.process_root_dir(executor, rootdir, tree_names, delete_existing_trees)
            file_futures = {}
            done = futures.wait(dir_futures, return_when=futures.FIRST_COMPLETED).done
            future = done.pop() if len(done) else None
            while future is not None:
                tree_path, level = dir_futures[future]
                del dir_futures[future]
                exc = future.exception()
                if exc is not None:
                    exc.tree_path = tree_path
                    raise exc

                processor_func = self.process_dir if not as_folder else self.process_folder
                dir_futures, file_futures = processor_func(executor, rootdir, tree_path, level, dir_futures,
                                                           file_futures)
                done = futures.wait(dir_futures, return_when=futures.FIRST_COMPLETED).done
                future = done.pop() if len(done) else None

            for future in futures.as_completed(file_futures):
                exc = future.exception()
                if exc is not None:
                    exc.tree_path = file_futures[future]
                    raise exc

    @staticmethod
    def _safe_makedirs(path):
        """makedirs if not exists"""
        try:
            makedirs(path)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise

    def create_node_on_filesystem(self, node, dirprefix, blob_dir):
        path = join(dirprefix, node['name'])
        if node['type'] == _Node.File:
            if node.get('blobRefId'):
                blob_path = join(blob_dir, node['blobRefId'])
                copyfile(blob_path, path)
            else:
                with open(path, 'w') as f:
                    f.write(node.get('data', ''))
        else:
            self._safe_makedirs(path)
            for child in node.get('children', []):
                self.create_node_on_filesystem(child, path, blob_dir)

    def download_tree(self, tree_name, rootdir, delete_existing, blob_temp_dir):
        if delete_existing:
            try:
                rmtree(join(rootdir, tree_name))
            except OSError as e:
                if e.errno != errno.ENOENT:
                    raise

        url = self._core_api_host + PARAMSERVER_API_TREE_PATH + tree_name
        response = RestClient(url).method(HttpMethod.GET).headers(self._headers).retry(0).execute()
        tree_root = get_api_response_data(response, parse_full=True).get('data', {})
        self.create_node_on_filesystem(tree_root, rootdir, blob_temp_dir)

    def get_blob_data(self, tree_names):
        url = self._core_api_host + PARAMSERVER_API_TREEBLOBS_PATH
        response = RestClient(url).method(HttpMethod.GET).query_param({'treeNames': tree_names}).headers(
            self._headers).retry(0).execute()
        blob_data = get_api_response_data(response, parse_full=True).get('data', {})
        return blob_data

    @staticmethod
    def download_blob_file(blob, blob_temp_dir):
        response = RestClient(blob['signedUrl']).method(HttpMethod.GET).execute()
        path = os.path.join(blob_temp_dir, str(blob['ID']))
        with open(path, 'wb') as f:
            for chunk in response.iter_content(1024 * 1024):
                f.write(chunk)

    @staticmethod
    def validate_args(rootdir, tree_names, delete_existing_trees, as_folder=False):
        if not isinstance(rootdir, six.string_types):
            raise InvalidParameterException('rootdir must be a string')
        if tree_names:
            validate_list_of_strings(tree_names, 'tree_names')
        if not isinstance(delete_existing_trees, bool):
            raise InvalidParameterException('delete_existing_trees must be a boolean')
        if not isinstance(as_folder, bool):
            raise InvalidParameterException('as_folder must be a boolean')

    def download_configurations(self, rootdir, tree_names, delete_existing_trees):
        self.validate_args(rootdir, tree_names, delete_existing_trees)
        self._safe_makedirs(rootdir)

        try:
            url = self._core_api_host + PARAMSERVER_API_TREE_PATH.rstrip('/')
            response = RestClient(url).method(HttpMethod.GET).headers(self._headers).retry(0).execute()
            api_tree_names = get_api_response_data(response, parse_full=True).get('data', [])
        except Exception as e:
            e.tree_path = ''
            raise

        if tree_names:
            api_tree_names = [tree_name for tree_name in api_tree_names if tree_name in tree_names]

        if not api_tree_names:
            raise ConfigNotFoundException('One or more trees not found')

        blob_temp_dir = tempfile.mkdtemp()

        blob_files = self.get_blob_data(api_tree_names)
        with futures.ThreadPoolExecutor(max_workers=15) as executor:
            blobs = blob_files.pop('blobRefs')
            blob_futures = {}
            for blob in blobs:
                if blob['signedUrl']:
                    future = executor.submit(self.download_blob_file, blob, blob_temp_dir)
                    blob_futures[future] = blob['ID']

            for future in futures.as_completed(blob_futures):
                exc = future.exception()
                if exc is not None:
                    # it is set to None for backward compatibility,
                    # the blob_files doesnt have any information abt tree
                    exc.tree_path = None
                    raise exc

        with futures.ThreadPoolExecutor(max_workers=15) as executor:
            tree_futures = {}
            for tree_name in api_tree_names:
                future = executor.submit(self.download_tree, tree_name, rootdir, delete_existing_trees, blob_temp_dir)
                tree_futures[future] = tree_name

            for future in futures.as_completed(tree_futures):
                exc = future.exception()
                if exc is not None:
                    exc.tree_path = tree_futures[future]
                    raise exc

        rmtree(blob_temp_dir)



from __future__ import absolute_import
from .clients.device import TopicKind, DeviceStatus, TopicQOS, QoS, DeploymentPhaseConstants, ROSDistro
from .clients.model import Label, Command, DeviceConfig, TopicsStatus
from rapyuta_io.utils import error
from .rio_client import Client
from .clients.device_manager import DeviceArch
from .clients.user_group import UserGroup

__version__ = "2.3.2"



# encoding: utf-8
from __future__ import absolute_import

import json
import random
import string
from functools import wraps

import requests
import six
import yaml
from six.moves import range

from rapyuta_io.utils import APIError, ParameterMissingException, InvalidParameterException, \
    UnauthorizedError, ResourceNotFoundError, BadRequestError, InternalServerError, ConflictError, \
    ForbiddenError, InvalidJSONError, InvalidYAMLError
from rapyuta_io.utils.settings import EMPTY, DEFAULT_RANDOM_VALUE_LENGTH

BEARER = "Bearer"


def prepend_bearer_to_auth_token(auth_token):
    if auth_token.startswith(BEARER):
        return auth_token
    else:
        return '{0} {1}'.format(BEARER, auth_token)


def create_auth_header(auth_token, project):
    return dict(Authorization=auth_token, project=project)


def get_error(response_data):
    try:
        err_response = json.loads(response_data)
        error_msg = err_response.get('error', None)
        if error_msg:
            return error_msg
        return err_response['response']['error']
    except (ValueError, TypeError, KeyError):
        return u''


def validate_key_value(obj):
    fields = ['id', 'key', 'value']
    for field in fields:
        field_value = obj.get(field, None)
        if field_value is None or field_value == '':
            raise ParameterMissingException(str.format("parameter {0} is missing", field))


def validate_list_of_strings(obj, parameter_name):
    if not isinstance(obj, list):
        raise InvalidParameterException('{0} must be a list'.format(parameter_name))
    for element in obj:
        if not isinstance(element, six.string_types):
            raise InvalidParameterException('Elements in {0} must be strings'.format(
                parameter_name))


def get_api_response_data(response, parse_full=False, errors=None, return_value=None):
    common_errors = {requests.codes.BAD_REQUEST: BadRequestError,
                     requests.codes.UNAUTHORIZED: UnauthorizedError,
                     requests.codes.NOT_FOUND: ResourceNotFoundError,
                     requests.codes.INTERNAL_SERVER_ERROR: InternalServerError,
                     requests.codes.CONFLICT: ConflictError,
                     requests.codes.FORBIDDEN: ForbiddenError}
    if errors:
        common_errors.update(errors)
    status_code = response.status_code
    if 200 <= status_code <= 299:
        if return_value:
            return return_value
        try:
            response_data = json.loads(response.text)
            if parse_full:
                return response_data
            else:
                return response_data['response']['data']
        except Exception as err:
            raise APIError(err)
    elif status_code in common_errors:
        raise common_errors[status_code](get_error(response.text))
    else:
        raise APIError(get_error(response.text))


def response_validator(parse_full=False, errors=None, return_value=None):
    def fun_wrapper(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            response = func(*args, **kwargs)
            return get_api_response_data(response, parse_full=parse_full,
                                         errors=errors, return_value=return_value)

        return wrapper

    return fun_wrapper


def write_to_file(file_name, content):
    file_ = open(file_name, 'w')
    file_.write(content)
    file_.close()
    return


def is_empty(obj):
    if obj is None or obj is EMPTY:
        return True
    return False


def remove_attributes(dict_obj, attr_list=list()):
    if dict_obj is not None:
        for attr in attr_list:
            dict_obj.pop(attr, None)
    return


def keep_attributes(dict_obj, attr_list=list()):
    for key in list(dict_obj):
        if key not in attr_list:
            dict_obj.pop(key, None)
    return


def generate_random_value(length=DEFAULT_RANDOM_VALUE_LENGTH):
    return ''.join(random.choice(string.ascii_lowercase) for _ in range(length))


def valid_list_elements(elements, type):
    if not isinstance(elements, list):
        return False
    return not [x for x in elements if not isinstance(x, type)]


def is_true(value):
    return value in [True, 'True', 'true']


def is_false(value):
    return value in [False, 'False', 'false']


def parse_json(filepath):
    """Parses the given file and checks if it is a valid JSON. If not, raises an error."""
    try:
        with open(filepath, 'r') as f:
            data = f.read()
    except Exception as e:
        raise e

    try:
        json.loads(data)
    except json.decoder.JSONDecodeError:
        raise InvalidJSONError(filepath)

    return data


def parse_yaml(filepath):
    """Parse the given file and checks if it is a valid YAML. If not, raises an error."""
    try:
        with open(filepath, 'r') as f:
            data = f.read()
    except Exception as e:
        raise e

    try:
        loaded = yaml.safe_load(data)
    except yaml.YAMLError:
        raise InvalidYAMLError(filepath)

    # For example, consider a file with just the following text.
    # The yaml.safe_load() function will still parse this file.
    #
    # invalid data
    #
    if not isinstance(loaded, dict):
        raise InvalidYAMLError(filepath)

    return data



from rapyuta_io.clients.project import Project, User
from rapyuta_io.utils import InvalidParameterException
from rapyuta_io.utils.object_converter import ObjBase, list_field


class UserGroup(ObjBase):
    """
    Usergroups in Rapyuta IO

    :ivar guid: Group ID
    :vartype guid: str
    :ivar name: Group Name
    :vartype name: str
    :ivar description: Description of the user group
    :vartype description: str
    :ivar creator: Creator's GUID
    :vartype creator: str
    :ivar members: Members who are part of the user group
    :vartype members: list(:py:class:`~rapyuta_io.clients.project.User`)
    :ivar admins: Admins of the user group
    :vartype admins: list(:py:class:`~rapyuta_io.clients.project.User`)
    :ivar projects: Projects that are part of the user group
    :vartype projects: list(:py:class:`~rapyuta_io.clients.project.Project`)
    """

    def __int__(self, guid=None, name=None, description=None,
                creator=None, members=None, admins=None, projects=None):

        self.validate(guid, name, description, creator, members, admins, projects)

        self.guid = guid
        self.name = name
        self.description = description
        self.creator = creator
        self.members = members
        self.admins = admins
        self.projects = projects
        self.role_in_projects = []

    @staticmethod
    def validate(self, guid, name, description, creator, members, admins, projects):
        if not isinstance(name, str) or not len(name):
            raise InvalidParameterException('name must be a non-empty string')

        if members and (not isinstance(members, list) or [m for m in members if not isinstance(m, User)]):
            raise InvalidParameterException('members must be a list of GroupMembers')

        if admins and (not isinstance(admins, list) or [a for a in admins if not isinstance(a, User)]):
            raise InvalidParameterException('members must be a list of GroupMembers')

        if projects and (not isinstance(projects, list) or [p for p in projects if not isinstance(p, Project)]):
            raise InvalidParameterException('members must be a list of GroupMembers')

    def get_serialize_map(self):
        return {
            'guid': 'guid',
            'name': 'name',
            'description': 'description',
            'creator': 'creator',
            'members': 'members',
            'admins': 'admins',
            'projects': 'projects',
            'userGroupRoleInProjects': 'role_in_projects',
        }

    def get_deserialize_map(self):
        return {
            'guid': 'guid',
            'name': 'name',
            'description': 'description',
            'creator': 'creator',
            'members': list_field('members', User),
            'admins': list_field('admins', User),
            'projects': list_field('projects', Project),
            'role_in_projects': 'userGroupRoleInProjects'
        }



# encoding: utf-8

from __future__ import absolute_import
from rapyuta_io.utils.error import OperationNotAllowedError
import six


class ObjDict(dict):
    def __contains__(self, k):
        try:
            return hasattr(self, k) or dict.__contains__(self, k)
        except:  # Not tested: in which situations can this happen?
            return False

    # only called if k not found in normal places
    def __getattr__(self, k):
        try:
            # Throws exception if not in prototype chain
            return object.__getattribute__(self, k)
        except AttributeError:
            try:
                return self[k]
            except KeyError:
                raise AttributeError(k)

    def __setattr__(self, k, v):
        try:
            # Throws exception if not in prototype chain
            object.__getattribute__(self, k)
        except AttributeError:
            try:
                self[k] = v
            except OperationNotAllowedError:
                raise OperationNotAllowedError
            except:  # Not tested: in which situations can this happen?
                raise AttributeError(k)
        else:  # Not tested: in which situations can this happen?
            object.__setattr__(self, k, v)

    def __delattr__(self, k):
        try:
            # Throws exception if not in prototype chain
            object.__getattribute__(self, k)
        except AttributeError:
            try:
                del self[k]
            except KeyError:
                raise AttributeError(k)
            except OperationNotAllowedError:
                raise OperationNotAllowedError
        else:  # Not tested: in which situations can this happen?
            object.__delattr__(self, k)

    def to_dict(self):
        return to_dict(self)

    @staticmethod
    def from_dict(d):
        return to_objdict(d)


class ImmutableKeyDict(ObjDict):
    def __setitem__(self, key, value):
        if key not in self and not (key == 'device_id'):
            raise OperationNotAllowedError
        super(ObjDict, self).__setitem__(key, value)

    def __delitem__(self, key):
        raise OperationNotAllowedError


def to_objdict(x):
    if isinstance(x, dict):
        return ObjDict((k, to_objdict(v)) for k, v in six.iteritems(x))
    elif isinstance(x, (list, tuple)):
        return type(x)(to_objdict(v) for v in x)
    else:
        return x


def to_dict(x):
    if isinstance(x, dict):
        return dict((k, to_dict(v)) for k, v in six.iteritems(x))
    elif isinstance(x, (list, tuple)):
        return type(x)(to_dict(v) for v in x)
    else:
        return x



# encoding: utf-8

default_host_config = {
    "core_api_host": "https://gaapiserver.apps.okd4v2.prod.rapyuta.io",
    "catalog_host": "https://gacatalog.apps.okd4v2.prod.rapyuta.io",
    "rip_host": "https://garip.apps.okd4v2.prod.rapyuta.io",
}

# Paramserver APIs
PARAMSERVER_API_BASE_PATH = '/api/paramserver/'
PARAMSERVER_API_TREE_PATH = PARAMSERVER_API_BASE_PATH + 'tree/'
PARAMSERVER_API_TREEBLOBS_PATH = PARAMSERVER_API_BASE_PATH + 'treeblobs'
PARAMSERVER_API_FILENODE_PATH = PARAMSERVER_API_BASE_PATH + 'filenode/'
# Device selection API host Config
SHARED_URL_PATH = '/logs/sharedurl'
DAEMONS_PATH = '/daemons'
DEVICE_API_BASE_PATH = '/api/device-manager/v0/'
DEVICE_PATH = 'device/'
DEVICE_API_PATH = DEVICE_API_BASE_PATH + 'devices/'
DEVICE_SELECTION_API_PATH = DEVICE_API_BASE_PATH + 'selection/query/'
DEVICE_CONFIG_VAR_API_PATH = DEVICE_API_BASE_PATH + 'config_variables/'
DEVICE_COMMAND_API_PATH = DEVICE_API_BASE_PATH + 'cmd/'
DEVICE_TOPIC_API_PATH = DEVICE_API_BASE_PATH + 'topics/'
PARAMETERS_API_PATH = DEVICE_API_BASE_PATH + 'parameters/'

# Metric
METRICS_API_BASE_PATH = '/api/metrics'
METRICS_API_QUERY_PATH = METRICS_API_BASE_PATH + '/v0/query/'
LIST_METRICS_API_QUERY_PATH = METRICS_API_BASE_PATH + '/v0/metrics/{}/{}'
LIST_TAGS_KEY_API_QUERY_PATH = METRICS_API_BASE_PATH + '/v0/tags/{}/{}'
LIST_TAGS_VALUE_API_QUERY_PATH = METRICS_API_BASE_PATH + '/v0/tags/{}/{}/{}'

GET_USER_PATH = '/api/user/me/get'

TOPIC_STATUS = '/status'
TOPIC_SUBSCRIBE = '/subscribe'
TOPIC_UNSUBSCRIBE = '/unsubscribed'

DEVICE_LABEL_API_PATH = DEVICE_API_BASE_PATH + 'labels/'
DEVICE_LOG_API_PATH = DEVICE_API_BASE_PATH + 'logs/'

FILE_DOWNLOAD_PATH = 'download?file_path='
DESERIALIZE_PATH = 'deserialize'
DEVICE_METRIC_API_PATH = DEVICE_API_BASE_PATH + 'metrics/'


# Device Onboarding Script
DEVICE_API_AUTH_KEY_PATH = DEVICE_API_BASE_PATH + 'auth-keys/{device_id}/token'
DEVICE_API_ADD_DEVICE_PATH = DEVICE_API_BASE_PATH + "auth-keys/?download_type=script"

# Catalog API Config
CATALOG_API_PATH = '/serviceclass/status'
PROVISION_API_PATH = '/v2/service_instances'
DEPROVISION_API_PATH = '/v2/service_instances'
SERVICE_INSTANCE_GUID = 'ser-guid'

INSTANCE_ID = 'instanceId'
ORGANIZATION_GUID = 'organizationGuid'
SPACE_GUID = 'spaceGuid'
BIND_ID = 'bind-rio'

VOLUME_PACKAGE_ID = 'io-public-persistent-volume'
LABEL_FORMAT = {
    "labels": {

        "select_criteria": {
            "fields": [
                {
                    "operator": "=",
                    "field": "",
                    "values": {"max_value": "", "min_value": ""}
                }
            ],
            "clause": "OR"
        }
    }
}

MEASUREMENT_FORMAT = {
    "measurements": {
        "measurement": "topicInt",
        "select_criteria": {
            "fields": [
                {
                    "operator": ">",
                    "field": "data",
                    "values": {"max_value": 700, "min_value": 100}
                }
            ],
            "clause": "OR"}
    }
}

EMPTY = ''
AUTH_TOKEN = ''
PROJECT = ''
DEVICE = 'device'
TOPICS = 'topics'
KIND = 'kind'
STATUS = 'status'
SUCCESS = 'success'
RESUMABLE_UPLOAD_URL = 'resumable_upload_url'
DEFAULT_SLEEP_INTERVAL = 6
DEPLOYMENT_STATUS_RETRY_COUNT = 50

BIND_ID_PREFIX = 'bind-'
DEFAULT_RANDOM_VALUE_LENGTH = 10

ENV_VAR_REGEX_PATTERN = r'[a-zA-Z_][a-zA-Z0-9_]*'



NON_PACKAGE_ATTRIBUTES = ['ID', 'id', 'name', 'ownerProject', 'CreatedAt', 'UpdatedAt',
                          'DeletedAt', 'guid', 'metadata']
NON_DEPLOYMENT_ATTRIBUTES = ['ID', 'creator', 'CreatedAt', 'UpdatedAt', 'DeletedAt', 'ownerProject'
                             'errors', 'dependentDeploymentStatus',
                             'packageDependencyStatus']
NON_PLAN_ATTRIBUTES = ['ID', 'id', 'DeletedAt', 'UpdatedAt', 'CreatedAt']

DEPLOYMENT_STATUS_ATTRIBUTES = ['deploymentId', 'planId', 'packageId', 'status', 'phase',
                                'errors', 'componentInfo', 'dependentDeploymentStatus',
                                'packageDependencyStatus']

NON_VOLUME_INSTANCE_ATTRIBUTES = ['ID', 'id', 'creator', 'ownerProject', 'CreatedAt',
                                  'UpdatedAt', 'DeletedAt', 'guid', 'metadata']

VOLUME_INSTANCE_STATUS_ATTRIBUTES = ['deploymentId', 'name', 'planId', 'packageId', 'status', 'phase',
                                     'errors', 'componentInfo', 'dependentDeploymentStatus',
                                     'packageDependencyStatus']

DEVICE = 'device'
LABELS = 'labels'
DEVICE_ID = 'device_id'



# encoding: utf-8
from __future__ import absolute_import
import json
import subprocess

from six.moves.urllib.parse import urlencode
from time import sleep
import enum

import requests
import six
import rapyuta_io

from rapyuta_io.clients.model import TopicsStatus, DeviceConfig, Label, Metric, LogUploadStatus, \
    LogUploads, SharedURL
from rapyuta_io.utils import ObjDict, RestClient, ParameterMissingException, \
    ConfigNotFoundException, to_objdict, LabelNotFoundException, DeviceNotFoundException, \
    DeploymentRunningException, \
    OperationNotAllowedError, UnknownTopicStatusException, LogsUUIDNotFoundException, \
    InvalidParameterException
from rapyuta_io.utils.rest_client import HttpMethod
from rapyuta_io.utils.settings import *
from rapyuta_io.utils.utils import create_auth_header, get_error, get_api_response_data, \
    validate_key_value, response_validator, is_true
from rapyuta_io.utils.partials import PartialMixin
from rapyuta_io.utils.pollers import RefreshPollerMixin

DEVICE_API_ERRORS = {
    400: ParameterMissingException,
    404: DeviceNotFoundException
}


class TopicQOS(enum.IntEnum):
    """
    .. deprecated:: 0.7.0
        Use :class:`QoS` instead.

        Enumeration variables for the Topic qos. Topic qos may be 0, 1 or 2 \n
        TopicQOS.ZERO \n
        TopicQOS.ONE \n
        TopicQOS.TWO \n
    """

    ZERO = 0
    ONE = 1
    TWO = 2


class QoS(enum.IntEnum):
    """
    Enumeration variables for QoS for topic/metric subscription. \n
    QoS.LOW \n
    QoS.MEDIUM \n
    QoS.HIGH \n
    """

    def __int__(self):
        return self.value

    LOW = 0
    MEDIUM = 1
    HIGH = 2


class TopicKind(str, enum.Enum):
    """
    Enumeration variables for the Topic kind. Topic kind may be 'Metric' or 'Log' \n
    TopicKind.METRIC \n
    TopicKind.LOG
    """

    def __str__(self):
        return str(self.value)

    METRIC = 'METRIC'
    LOG = 'LOG'


class DeviceStatus(str, enum.Enum):
    """
    DeviceStatus enumeration represents the supported device status.
    Device status can be any of the below types \n
    DeviceStatus.ONLINE \n
    DeviceStatus.REJECTED \n
    DeviceStatus.ACCEPTED \n
    DeviceStatus.OFFLINE \n
    DeviceStatus.REGISTERED \n
    DeviceStatus.INITIALIZING \n
    DeviceStatus.FAILED \n
    DeviceStatus.NEW \n
    DeviceStatus.DELETED \n
    """

    def __str__(self):
        return str(self.value)

    ONLINE = 'ONLINE'
    REJECTED = 'REJECTED'
    ACCEPTED = 'ACCEPTED'
    OFFLINE = 'OFFLINE'
    REGISTERED = 'REGISTERED'
    INITIALIZING = 'INITIALIZING'
    FAILED = 'FAILED'
    NEW = 'NEW'
    DELETED = 'DELETED'


class SystemMetric(str, enum.Enum):
    """
    Metrics options that are currently supported by device.
    User can subscribe any of metrics options given below \n
    SystemMetric.CPU \n
    SystemMetric.MEMORY \n
    SystemMetric.DISK \n
    SystemMetric.DISKIO \n
    SystemMetric.NETWORK \n
    SystemMetric.WIRELESS \n
    """

    def __str__(self):
        return str(self.value)

    CPU = 'cpu'
    MEMORY = 'memory'
    DISK = 'disk'
    DISKIO = 'diskio'
    NETWORK = 'network'
    WIRELESS = 'wireless'


class DeviceRuntime(str, enum.Enum):
    """
    Enumeration variables for the Device runtimes available in rapyuta.io. \n
    DeviceRuntime.DOCKER ('dockercompose') \n
    DeviceRuntime.PREINSTALLED ('preinstalled')
    """

    def __str__(self):
        return str(self.value)

    DOCKER = 'dockercompose'
    PREINSTALLED = 'preinstalled'


class DeviceRuntimeKeys(str, enum.Enum):
    """
    The Device runtime keys to enable/disable runtimes available in rapyuta.io. \n
    DeviceRuntimeKeys.RUNTIME_DOCKER ('runtime_docker') \n
    DeviceRuntimeKeys.RUNTIME_PREINSTALLED ('runtime_preinstalled')
    """

    def __str__(self):
        return str(self.value)

    RUNTIME_DOCKER = 'runtime_docker'
    RUNTIME_PREINSTALLED = 'runtime_preinstalled'


class OnboardScript(object):
    """
    OnboardScript class can be used to download and run the device onboard script.

    :ivar token: Authorization token needed to make api call
    :ivar url: URL to download the onboard script
    :ivar command: command to execute the downloaded script

    """

    def __init__(self, token, url, command):
        self.token = token
        self.url = url
        self.command = command

    def run(self):
        """
        Downloads the onboard script and executes it in a subprocess. In case the script
        returns non-zero code after execution, an exception is raised.
        """
        # signal module gets imported only when the run method gets called to
        # avoid its redundant import everytime user imports rapyuta_io. Otherwise,
        # this might lead to implications on signal handling in user’s application.
        # https://stackoverflow.com/q/39397034
        import signal
        signal.signal(signal.SIGPIPE, signal.SIG_DFL)
        headers = dict()
        headers['Authorization'] = 'Bearer ' + self.token
        response = RestClient(self.url).method(HttpMethod.GET).headers(headers).execute()
        response = get_api_response_data(response, return_value=response)
        with open('start', 'w') as f:
            f.write(response.text)
        command = self.command.split()
        return_code = subprocess.call(command)
        if return_code != 0:
            raise Exception('Command {} returned with non-zero exit code'.format(self.command))

    def full_command(self):
        """
        Returns the full bash command that downloads and executes the onboard script.

        :return: returns the full bash command to onboard a device.
        :rtype: str
        """
        return "curl -O -H 'Authorization: Bearer {token}' {url} && {command}".format(token=self.token, url=self.url,
                                                                                      command=self.command)


class DevicePythonVersion(str, enum.Enum):
    """
    Enumneration variables for the Device python version available in rapyuta.io \n
    DevicePythonVersion.PYTHON2 ('2') \n
    DevicePythonVersion.PYTHON3 ('3')
    """

    def __str__(self):
        return str(self.value)

    PYTHON2 = '2'
    PYTHON3 = '3'


class Device(PartialMixin, RefreshPollerMixin, ObjDict):
    """
    Device class represents a device. Member variables of the class represent the
    properties of device. \n
    Variables marked as (full-only) are only available on a full object. Use `refresh()` to convert a
    partial object into a full one.

    :ivar uuid: Id of the device.
    :ivar name: Name of the device.
    :ivar status: Status of the device.
    :vartype status: :py:class:`~rapyuta_io.clients.device.DeviceStatus`
    :ivar username: (full-only) User log in to device
    :ivar saltversion: (full-only) Salt version of the device.
    :ivar registration_time: Device registration time.
    :ivar last_online: Time when device was last online.
    :ivar python_version: indicates whether device is running on python2 or python3
    :ivar description: Description of the device.
    :ivar labels: List of labels associated with the device.
    :ivar config_variables: Configuration variables of the device.
    :ivar deployments: List of deployments on device
    :ivar error_code: Error code when device goes to FAILED state. Will be of the form: 'DEV_E*'.
    :ivar error_message: Error message when device goes to FAILED state.
    :ivar created_by: Creator user guid.
    :ivar host: (full-only) Hostname on device.
    :ivar ip_interfaces: (full-only) Set of IP Interfaces on device.
    :ivar lsb_distrib_description: (full-only) LSB Distribution details.

    """

    RUNTIME = 'runtime'
    PRE_INSTALLED = 'preinstalled'
    DOCKER_COMPOSE = 'dockercompose'

    def __init__(self, name, runtime=None, runtime_docker=False, runtime_preinstalled=False, ros_distro=None,
                 ros_workspace=None, description=None, python_version=DevicePythonVersion.PYTHON2,
                 config_variables=None, labels=None):
        self.validate(name, runtime, runtime_docker, runtime_preinstalled, ros_distro, ros_workspace,
                      description, python_version)
        self.name = name
        # The below is done to ensure backward compatibility
        if runtime == DeviceRuntime.DOCKER:
            self._runtime_docker = True
            self._runtime_preinstalled = False
        elif runtime == DeviceRuntime.PREINSTALLED:
            self._runtime_docker = False
            self._runtime_preinstalled = True
        else:
            self._runtime_docker = runtime_docker
            self._runtime_preinstalled = runtime_preinstalled
        self._ros_distro = ros_distro
        self._ros_workspace = ros_workspace
        self.description = description
        self.python_version = python_version
        self.config_variables = config_variables or {}
        self.labels = labels or {}

    @staticmethod
    def validate(name, runtime, runtime_docker, runtime_preinstalled, ros_distro,
                 ros_workspace, description, python_version):
        if not name or not isinstance(name, six.string_types):
            raise InvalidParameterException('name must be a non-empty string')
        if runtime is not None and (runtime not in list(DeviceRuntime.__members__.values())):
            raise InvalidParameterException('runtime must be one of rapyuta_io.clients.device.DeviceRuntime')
        if not isinstance(runtime_docker, bool):
            raise InvalidParameterException('runtime_docker must be a boolean')
        if not isinstance(runtime_preinstalled, bool):
            raise InvalidParameterException('runtime_preinstalled must be a boolean')
        if python_version is not None and (
                python_version not in list(DevicePythonVersion.__members__.values())):
            raise InvalidParameterException(
                'python_version must be one of rapyuta.io.client.device.DevicePythonVersion')
        if ros_distro is not None and (
                ros_distro not in list(ROSDistro.__members__.values())):
            raise InvalidParameterException('ros_distro must be one of rapyuta_io.clients.package.ROSDistro')
        if runtime_preinstalled and ros_distro == ROSDistro.NOETIC:
            raise InvalidParameterException('preinstalled runtime does not support noetic ros_distro yet')
        if ros_distro == ROSDistro.NOETIC and python_version == DevicePythonVersion.PYTHON2:
            raise InvalidParameterException('noetic ros_distro not supported on python_version 2')
        if ros_workspace is not None and not isinstance(ros_workspace, six.string_types):
            raise InvalidParameterException('ros_workspace must be of type string')
        if description is not None and not isinstance(description, six.string_types):
            raise InvalidParameterException('description must be of type string')

    def _execute_api(self, url, request_method=HttpMethod.GET, payload=None, retry_limit=0, query=None):
        headers = create_auth_header(self._auth_token, self._project)
        headers['Content-Type'] = 'application/json'
        rest_client = RestClient(url).method(request_method).headers(headers).query_param(query)
        response = rest_client.retry(retry_limit).execute(payload=payload)
        return response

    def _serialize(self):
        device = dict()
        device['name'] = self.name
        device['python_version'] = self.python_version
        device['config_variables'] = {}
        if self.description:
            device['description'] = self.description
        for field in ['_runtime_docker', '_runtime_preinstalled', '_ros_distro','_ros_workspace']:
            if getattr(self, field):
                device['config_variables'][field[1:]] = getattr(self, field)
        device['config_variables'].update(self.config_variables)
        device['labels'] = self.labels
        return device

    @classmethod
    def _deserialize(cls, data):
        obj = cls.__new__(cls)
        for key, value in six.iteritems(data):
            # putting this hack since python_version makes more sense from readability perspective and device_version
            # is more specific to backend implementation
            if key == 'device_version':
                setattr(obj, 'python_version', DevicePythonVersion(value))
                continue
            setattr(obj, key, value)

        setattr(obj, 'deviceId', obj.uuid)
        obj.deployments = []
        if obj.config_variables:
            obj.config_variables = [DeviceConfig(to_objdict(config)) for config in obj.config_variables]
        if obj.labels:
            obj.labels = [Label(to_objdict(label)) for label in obj.labels]
        return obj

    def is_online(self):
        if self.status == DeviceStatus.ONLINE.value:
            return True
        return False

    def is_ready(self):
        return self.is_online()

    def get_runtime(self):
        """
        Get the device runtime
        :return: value of device configuration variable: runtime

        Following example demonstrates how to get device runtime.

        >>> from rapyuta_io import Client
        >>> client = Client(auth_token='auth_token', project="project_guid")
        >>> device = client.get_device('test_device_id')
        >>> device.get_runtime()
        """
        for config in self.config_variables:
            if config.key == self.RUNTIME:
                return config.value

    def is_docker_enabled(self):
        """
        Check if docker runtime is enabled
        :return: boolean value depicting if docker runtime is enabled/disabled on the device

        Following example demonstrates how to check if docker runtime is enabled/disabled on the device.

        >>> from rapyuta_io import Client
        >>> client = Client(auth_token='auth_token', project="project_guid")
        >>> device = client.get_device('test_device_id')
        >>> device.is_docker_enabled()
        """
        for config in self.config_variables:
            if config.key != DeviceRuntimeKeys.RUNTIME_DOCKER:
                continue

            return is_true(config.value)

        return False

    def is_preinstalled_enabled(self):
        """
        Check if preinstalled runtime is enabled
        :return: boolean value depicting if preinstalled runtime is enabled/disabled on the device

        Following example demonstrates how to check if preinstalled runtime is enabled/disabled on the device.

        >>> from rapyuta_io import Client
        >>> client = Client(auth_token='auth_token', project="project_guid")
        >>> device = client.get_device('test_device_id')
        >>> device.is_preinstalled_enabled()
        """
        for config in self.config_variables:
            if config.key != DeviceRuntimeKeys.RUNTIME_PREINSTALLED:
                continue

            return is_true(config.value)

        return False

    def onboard_script(self):
        """
        Returns the OnboardScript object. This object has the information regarding api server url,
        auth token and command to be used to install rapyuta-agent on the device. The run method of
        this object is supposed to be called in order to install rapyuta-agent on the device.

        :return: returns the OnboardScript object
        :rtype: :py:class:`~rapyuta_io.clients.device.OnboardScript`

        Following example demonstrates how to use the onboard_script method.

        >>> from rapyuta_io import Client
        >>> client = Client(auth_token='auth_token', project="project_guid")
        >>> device = client.get_device('test_device_id')
        >>> onboard_script = device.onboard_script()
        >>> onboard_script.run()
        """
        url = self._device_api_host + DEVICE_API_AUTH_KEY_PATH.format(device_id=self.deviceId)
        response = self._execute_api(url, HttpMethod.GET)
        response_data = get_api_response_data(response, parse_full=True)
        token = response_data['response']['data']
        url = self._device_api_host + '/start'
        command = response_data['response']['script_command']
        return OnboardScript(token, url, command)

    @response_validator(errors=DEVICE_API_ERRORS, return_value=True)
    def save(self, retry_limit=0):
        """
        Update device details such as device name, device status and description.

        :param retry_limit: Optional parameter to specify the number of retry attempts to be
               carried out if any failures occurs during the API call.
        :type retry_limit: int
        :return: returns true if the device details updated successfully.
        :raises: :py:class:`~utils.error.DeviceNotFountException`:  If the device is not found.
        :raises: :py:class:`~utils.error.ParameterMissingException`:  If any parameters are
                 missing in the request.
        :raises: :py:class:`~utils.error.APIError`: If the api returns an error, the status code is
            anything other than 200/201

        Following example demonstrates how to update a device.

        >>> from rapyuta_io import Client
        >>> client = Client(auth_token='auth_token', project="project_guid")
        >>> device = client.get_device('test_device_id')
        >>> device.name = 'new_device_name'
        >>> device.save()

        """
        url = self._device_api_host + DEVICE_API_PATH
        device = self
        # todo: update backend api to get id in url
        device.device_id = self.uuid
        return self._execute_api(url, HttpMethod.PUT, self, retry_limit)

    def refresh(self, retry_limit=0):
        """
        Fetches the updated resource from the server, and adds/updates object attributes based on it.

        :param retry_limit: Optional parameter to specify the number of retry attempts to be
              carried out if any failures occurs during the API call.
        :type retry_limit: int
        :raises: :py:class:`~utils.error.DeviceNotFountException`:  If the device is not found.
        :raises: :py:class:`~utils.error.APIError`: If the api returns an error, the status code is
            anything other than 200/201

        """
        url = self._device_api_host + DEVICE_API_PATH + self.uuid
        response = self._execute_api(url, HttpMethod.GET, retry_limit=retry_limit)
        device_data = get_api_response_data(response)
        device = Device._deserialize(device_data)
        for attr in device.keys():
            self.__setattr__(attr, device.__getattr__(attr))
        self.is_partial = False

    def reject_device(self, retry_limit=0):
        self.status = DeviceStatus.REJECTED
        return self.save(retry_limit)

    def accept_device(self, retry_limit=0):
        self.status = DeviceStatus.ACCEPTED
        return self.save(retry_limit)

    def delete(self, retry_limit=0):
        """
        Delete the device

        :param retry_limit: Optional parameter to specify the number of retry attempts to be
              carried out if any failures occur during the API call.
        :type retry_limit: int
        :return: True if the device is deleted successfully. Otherwise returns False.
        :raises: :py:class:`~utils.error.DeviceNotFoundException`:  If the device is not found.
        :raises: :py:class:`~utils.error.DeploymentRunningException`:  When any deployment is
                 running on the device.
        :raises: :py:class:`~utils.error.APIError`: If the api returns an error, the status code
            is anything other than 200/201

        Following example demonstrates how to delete a device

         >>> from rapyuta_io import Client
         >>> from rapyuta_io.clients.model import Label
         >>> client = Client(auth_token='auth_token', project="project_guid")
         >>> device = client.get_device('test_device_id')
         >>> device.delete()

        """

        for deployment in self.deployments:
            if deployment.phase.lower() == DeploymentPhaseConstants.SUCCEEDED.value.lower():
                raise DeploymentRunningException('Cannot delete device. Deployment %s is running '
                                                 'on the device' % deployment.deployment_id)

        url = self._device_api_host + DEVICE_API_PATH + self.uuid
        headers = create_auth_header(self._auth_token, self._project)
        response = RestClient(url).method(HttpMethod.DELETE).headers(headers) \
            .retry(retry_limit).execute()
        if response.status_code == requests.codes.BAD_REQUEST:
            raise DeploymentRunningException()
        delete_status = get_api_response_data(response, True)
        if delete_status[STATUS] == SUCCESS:
            self.clear()
            return True
        return False

    def execute_command(self, command, retry_limit=0):
        """
        Execute command on device

        :param command: Command to execute
        :type command: :py:class:`~rapyuta_io.clients.model.Command`
        :param retry_limit: Optional parameter to specify the number of retry attempts to be
              carried out if any failures occurs during the API call.
        :type retry_limit: int
        :return: Execution result

        :raises: :py:class:`~utils.error.ParameterMissingException`: If any parameters are missing
                 in the request.
        :raises: :py:class:`~utils.error.APIError`: If the api returns an error, the status code is
            anything other than 200/201

        Following example demonstrates how to execute a command.

         >>> from rapyuta_io import Client
         >>> from rapyuta_io.clients.model import Command
         >>> client = Client(auth_token='auth_token', project="project_guid")
         >>> device = client.get_device('test_device_id', )
         >>> command = Command(cmd='uname -a', shell='/bin/bash', bg=False)
         >>> device.execute_command(command)

        """
        command.validate()
        command.device_ids = [self.uuid]
        url = self._device_api_host + DEVICE_COMMAND_API_PATH
        response = self._execute_api(url, HttpMethod.POST, command.to_json(), retry_limit)
        if response.status_code == requests.codes.BAD_REQUEST:
            raise ParameterMissingException(get_error(response.text))
        execution_result = get_api_response_data(response)
        if not command.run_async:
            return execution_result 
        
        jid = execution_result.get('jid')
        if not jid:
            raise ValueError("Job ID not found in the response")
        return self.fetch_command_result(jid, [self.uuid], timeout=command.timeout)

    def fetch_command_result(self, jid: str, deviceids: list, timeout: int, interval: int = 10):
        """
        Fetch the result of the command execution using the job ID (jid) and the first device ID from the list.
        Args:
            jid (str): The job ID of the executed command.
            deviceids (list): A list of device IDs on which the command was executed.
            timeout (int): The maximum time to wait for the result (in seconds). Default is 300 seconds.
            interval (int): time interval for retry
        Returns:
            dict: The result of the command execution.
        Raises:
            TimeoutError: If the result is not available within the timeout period.
            APIError: If the API returns an error.
        """

        if not deviceids or not isinstance(deviceids, list):
            raise ValueError("Device IDs must be provided as a non-empty list.")
        url = self._device_api_host + DEVICE_COMMAND_API_PATH + jid
        query = {
            "jid": jid,
            "device_id": deviceids[0]
        }
        time_elapsed = 0
        wait_interval = interval
        while time_elapsed < timeout:
            response = self._execute_api(url, HttpMethod.GET, query=query)
            if response.status_code == requests.codes.OK:
                result = get_api_response_data(response)
                return result[deviceids[0]]
            sleep(wait_interval)
            time_elapsed += wait_interval
        raise TimeoutError(f"Command result not available after {timeout} seconds")

    def get_config_variables(self):
        """
        Get configuration variables associated with the device

        :return: list of instances of :py:class:`~clients.DeviceConfig` class.
        :raises: :py:class:`~utils.error.APIError`: If the api returns an error, the status code is
            anything other than 200/201

        """
        url = self._device_api_host + DEVICE_CONFIG_VAR_API_PATH + DEVICE_PATH + self.uuid
        response = self._execute_api(url, HttpMethod.GET)
        config_list = get_api_response_data(response)
        config_vars = [DeviceConfig(to_objdict(config)) for config in config_list]
        self.config_variables = config_vars
        return self.config_variables

    def add_config_variable(self, key, value):
        """
        Add new configuration variable to the device.

        :param key: Configuration variable key
        :type key: string
        :param value: Configuration value
        :type value: string
        :return: instance of :py:class:`~.DeviceConfig` class.
        :raises: :py:class:`~utils.error.ParameterMissingException`: If any parameters are missing
                 in the request.
        :raises: :py:class:`~utils.error. DeviceNotFoundException`: If the device is not found.
        :raises: :py:class:`~utils.error.APIError`: If the API returns an error, the status code is
            anything other than 200/201

        Following example demonstrates how to add a device configuration variable.

         >>> from rapyuta_io import Client
         >>> client = Client(auth_token='auth_token', project="project_guid")
         >>> device = client.get_device('test_device_id')
         >>> device.add_config_variable('config_key', 'config_value')

        """

        if not key or not value:
            raise ParameterMissingException()
        config_var = {'key': key, 'value': value}
        url = self._device_api_host + DEVICE_CONFIG_VAR_API_PATH + DEVICE_PATH + self.uuid
        response = self._execute_api(url, HttpMethod.POST, payload=config_var)
        if response.status_code == requests.codes.BAD_REQUEST:
            raise ParameterMissingException(get_error(response.text))
        config_var = get_api_response_data(response)
        config_obj = DeviceConfig(to_objdict(config_var))
        self.config_variables.append(config_obj)
        return config_obj

    def update_config_variable(self, config):
        """
        Update device configuration variable of a device.

        :param config: Configuration variable
        :type config: instance of :py:class:`~.DeviceConfig` class
        :return: instance of :py:class:`~.DeviceConfig` class.
        :raises: :py:class:`~utils.error.ParameterMissingException`: If any parameters are missing
                 in the request.
        :raises: :py:class:`~utils.error. DeviceNotFoundException`: If the device is not found.
        :raises: :py:class:`~utils.error.APIError`: If the API call returns an error, the status code
            is anything other than 200/201.

        Following example demonstrates how to update a device configuration variable.

         >>> from rapyuta_io import Client
         >>> client = Client(auth_token='auth_token', project="project_guid")
         >>> device = client.get_device('test_device_id')
         >>> config_var = device.get_config_variables()[0]
         >>> config_var.value = 'new_config_value'
         >>> device.update_config_variable(config_var)

        """

        validate_key_value(config)
        if not config.is_updatable():
            raise OperationNotAllowedError('Default config variable {} cannot be '
                                           'updated'.format(config.key))

        url = self._device_api_host + DEVICE_CONFIG_VAR_API_PATH + str(config.id)
        response = self._execute_api(url, HttpMethod.PUT, payload=config)
        if response.status_code == requests.codes.BAD_REQUEST:
            raise ParameterMissingException(get_error(response.text))
        config_var = get_api_response_data(response)
        for config in self.config_variables:
            if config.id == config_var['id']:
                config.value = config_var['value']
                return config
        return DeviceConfig(to_objdict(config_var))

    def delete_config_variable(self, config_id, retry_limit=0):
        """
        Delete configuration variable for the device.

        :param config_id: Configuration variable id
        :type config_id: int
        :param retry_limit: Optional parameter to specify the number of retry attempts to be
              carried out if any failures occur during the API call.
        :type retry_limit: int
        :return: True if the configuration variable is deleted, otherwise returns false
        :raises: :py:class:`~utils.error.ParameterMissingException`: If any parameters are missing
                 in the request.
        :raises: :py:class:`~utils.error. ConfigNotFoundException`: If the configuration
                 is not found.
        :raises: :py:class:`~utils.error.APIError`: If the api returns an error, the status code is
            anything other than 200/201

        Following example demonstrates how to delete a device configuration variable.

         >>> from rapyuta_io import Client
         >>> client = Client(auth_token='auth_token', project="project_guid")
         >>> device = client.get_device('test_device_id')
         >>> device.delete_config_variable(config_id='config_id')

        """
        for config in self.config_variables:
            if config.id == config_id and not config.is_deletable():
                raise OperationNotAllowedError('Default config variable {} cannot be '
                                               'deleted'.format(config.key))

        url = self._device_api_host + DEVICE_CONFIG_VAR_API_PATH + str(config_id)
        response = self._execute_api(url, HttpMethod.DELETE, retry_limit)
        if response.status_code == requests.codes.NOT_FOUND:
            raise ConfigNotFoundException(get_error(response.text))
        delete_status = get_api_response_data(response, True)
        if delete_status[STATUS] == SUCCESS:
            for config in self.config_variables:
                if config.id == config_id:
                    self.config_variables.remove(config)
                    break
            return True
        return False

    def get_deployments(self):
        """
        Get partial details of deployments associated with the device. Also, update `deployments` field with these. \n

        **Note**: Deployment details are partial. Use :meth:`~rapyuta_io.clients.deployment.Deployment.refresh` as shown
        in the example below to refresh these to full objects.

        :return: List of deployment details.

        Following example demonstrates how to get device deployments

         >>> from rapyuta_io import Client
         >>> client = Client(auth_token='auth_token', project="project_guid")
         >>> device = client.get_device('test_device_id')
         >>> partial_deployments = device.get_deployments()
         >>> # to get full deployments objects:
         >>> for deployment in partial_deployments:
         >>>     deployment.refresh()

        """
        url = self._device_api_host + DEVICE_API_PATH + self.deviceId
        response = self._execute_api(url, HttpMethod.GET)
        if response.status_code == requests.codes.NOT_FOUND:
            raise DeviceNotFoundException(get_error(response.text))
        device_resp = get_api_response_data(response)
        deployment_list = device_resp['deployments']
        self.deployments = [to_objdict(dep) for dep in deployment_list]
        return self.deployments

    def get_labels(self):
        """
        Get all labels associated with the device

        :return: list of instances of the class :py:class:`~clients.model.Label`.

        Following example demonstrates how to get device labels

         >>> from rapyuta_io import Client
         >>> client = Client(auth_token='auth_token', project="project_guid")
         >>> device = client.get_device('test_device_id')
         >>> device.get_labels()

        """

        url = self._device_api_host + DEVICE_LABEL_API_PATH + self.uuid
        response = self._execute_api(url, HttpMethod.GET)
        if response.status_code == requests.codes.NOT_FOUND:
            raise DeviceNotFoundException(get_error(response.text))
        self.labels = [Label(to_objdict(label)) for label in get_api_response_data(response)]
        return self.labels

    def add_label(self, key, value, retry_limit=0):
        """
        Add label to the device

        :param key: label key
        :type key: string
        :param key: label value
        :type key: string
        :param retry_limit: Optional parameter to specify the number of retry attempts to be carried out if any failures occur during the API call.
        :type retry_limit: int
        :return: instances of the class :py:class:`~clients.model.Label`.
        :raises: :py:class:`~utils.error. DeviceNotFoundException`: If the device is not found.
        :raises: :py:class:`~utils.error.ParameterMissingException`: If any parameters are missing in the request.
        :raises: :py:class:`~utils.error.APIError`: If the API call returns an error, the status code
            is anything other than 200/201

        Following example demonstrates how to add device labels

         >>> from rapyuta_io import Client
         >>> from rapyuta_io.clients.model import Label
         >>> client = Client(auth_token='auth_token', project="project_guid")
         >>> device = client.get_device('test_device_id')
         >>> device.add_label('key', 'value')

        """

        request_data = dict()
        if not key or not value:
            raise ParameterMissingException()
        request_data[key] = value
        url = self._device_api_host + DEVICE_LABEL_API_PATH + self.uuid
        response = self._execute_api(url, HttpMethod.POST, request_data, retry_limit)
        if response.status_code == requests.codes.BAD_REQUEST:
            raise ParameterMissingException(get_error(response.text))
        label = Label(to_objdict(get_api_response_data(response)[0]))
        self.labels.append(label)
        return label

    def update_label(self, label, retry_limit=0):
        """
        Update device label

        :param label: instance of class :py:class:`~clients.model.Label` to update
        :type label: :py:class:`~clients.model.Label`
        :param retry_limit: Optional parameter to specify the number of retry attempts to be
              carried out if any failures occurs during the API call.
        :type retry_limit: int
        :return: updated instance of the class :py:class:`Label`.
        :raises: :py:class:`~utils.error.LabelNotFoundException`: If the label is not found.
        :raises: :py:class:`~utils.error.APIError`: If the API call returns an error, the status code
            is anything other than 200/201.

        Following example demonstrates how to update a device label

         >>> from rapyuta_io import Client
         >>> from rapyuta_io.clients.model import Label
         >>> client = Client(auth_token='auth_token', project="project_guid")
         >>> device = client.get_device('test_device_id')
         >>> label = device.get_labels()[0]
         >>> label.value = 'new label value'
         >>> device.update_label(label)

        """
        validate_key_value(label)
        url = self._device_api_host + DEVICE_LABEL_API_PATH + str(label.id)
        response = self._execute_api(url, HttpMethod.PUT, label, retry_limit)
        if response.status_code == requests.codes.NOT_FOUND:
            raise LabelNotFoundException()
        for _label in self.labels:
            if _label.id == label.id:
                _label.key = label.key
                _label.value = label.value
                return _label
        return label

    def delete_label(self, label_id, retry_limit=0):
        """
        Delete a device label.

        :param label_id: Label Id to delete
        :type label_id: int
        :param retry_limit: Optional parameter to specify the number of retry attempts to be
              carried out if any failures occurs during the API call.
        :type retry_limit: int
        :return: Boolean value to indicate whether the label is deleted or not
        :raises: :py:class:`~utils.error.LabelNotFoundException`: If the label is not found.
        :raises: :py:class:`~utils.error.APIError`: If the API call returns an error, the status code
            is anything other than 200/201.

        Following example demonstrates how to delete a device label

         >>> from rapyuta_io import Client
         >>> client = Client(auth_token='auth_token', project="project_guid")
         >>> device = client.get_device('test_device_id')
         >>> device.delete_label('label_id')

        """

        url = self._device_api_host + DEVICE_LABEL_API_PATH + str(label_id)
        response = self._execute_api(url, HttpMethod.DELETE, retry_limit=retry_limit)
        if response.status_code == requests.codes.NOT_FOUND:
            raise LabelNotFoundException()
        delete_status = get_api_response_data(response, True)
        if delete_status[STATUS] == SUCCESS:
            for label in self.labels:
                if label.id == label_id:
                    self.labels.remove(label)
                    break
            return True
        return False

    def topics(self, retry_limit=0):
        """
        Fetching the available topics on the device

        :param retry_limit: Optional parameter to specify the number of retry attempts to be
              carried out if any failures occurs during the API call.
        :type retry_limit: int
        :return: List of topics in the device (list(string)
        :raises: :py:class:`~utils.error. DeviceNotFountException`: If the device is not found.
        :raises: :py:class:`~utils.error.APIError`: If the api returns an error, the status code is
            anything other than 200/201

        Following example demonstrates how to get the topic list

                >>> from rapyuta_io import Client
                >>> client = Client(auth_token='auth_token', project="project_guid")
                >>> device = client.get_device('test_device_id')
                >>> device.topics()

        """

        url = self._device_api_host + DEVICE_TOPIC_API_PATH + self.uuid
        response = self._execute_api(url, retry_limit=retry_limit)
        topic_info = get_api_response_data(response)
        if isinstance(topic_info, list):
            if topic_info[0] == requests.codes.OK:
                if isinstance(topic_info[1], dict):
                    return topic_info[1][TOPICS]
                return json.loads(str(topic_info[1]))[TOPICS]

        return list()

    def topic_status(self, retry_limit=0):
        """
        Get the subscribed and unsubscribed topics status

        :param retry_limit: Optional parameter to specify the number of retry attempts to be
              carried out if any failures occurs during the API call.
        :type retry_limit: int
        :return: Instance of class :py:class:`~clients.model.TopicsStatus`.
        :raises: :py:class:`~utils.error. DeviceNotFoundException`: If the device is not found.
        :raises: :py:class:`~utils.error. UnknownTopicStatusException`: If the topic status is empty.
        :raises: :py:class:`~utils.error.APIError`: If the API call returns an error, the status code is
            anything other than 200/201.

        Following example demonstrates how to get the topic status

         >>> from rapyuta_io import Client
         >>> client = Client(auth_token='auth_token', project="project_guid")
         >>> device = client.get_device('test_device_id')
         >>> device.topic_status()

        """

        url = self._device_api_host + DEVICE_TOPIC_API_PATH + self.uuid + TOPIC_STATUS
        response = self._execute_api(url, retry_limit=retry_limit)
        topic_status = get_api_response_data(response)

        if not topic_status:
            raise UnknownTopicStatusException('No topics were found')
        topics_info = get_api_response_data(response)[1]

        if not isinstance(topics_info, dict):
            topics_info = json.loads(topics_info)

        topic_status = TopicsStatus(to_objdict(topics_info[TOPICS]))
        topic_status.master_up = topics_info['master_up']
        return topic_status

    def subscribe_topic(self, topic, qos=QoS.LOW.value, kind=TopicKind.METRIC, whitelist_tag=[],
                        whitelist_field=[], fail_on_topic_inexistence=True, retry_limit=0):
        """

        To subscribe a topic

        :param topic: topic name to subscribe
        :type topic: string
        :param qos: QoS value for topic subscription
        :type qos: integer
        :param kind: subscribe the topic as a metric or log
        :type kind: enum :py:class:`~TopicKind`
        :param whitelist_tag: Optional parameter to specify the tags that needs to be whitelisted in metrics
        :type whitelist_tag: list
        :param whitelist_field: Optional parameter to specify the fields that needs to be whitelisted in metrics
        :type whitelist_field: list
        :param fail_on_topic_inexistence: Optional parameter to fail with error if topic doesnt exists on device
        :type fail_on_topic_inexistence: bool
        :param retry_limit: Optional parameter to specify the number of retry attempts to be
              carried out if any failures occurs during the API call.
        :type retry_limit: int
        :return: list containing subscription status.
        :raises: :py:class:`~utils.error. DeviceNotFountException`: If the device is not found.
        :raises: :py:class:`~utils.error.APIError`: If the api returns an error, the status code is
            anything other than 200/201

        Following example demonstrates how to subscribe a topic

         >>> from rapyuta_io import Client
         >>> from rapyuta_io.clients.device import QoS
         >>> client = Client(auth_token='auth_token', project="project_guid")
         >>> device = client.get_device('test_device_id')
         >>> device.subscribe_topic('test_topic', qos=QoS.MEDIUM.value, kind=TopicKind.METRIC)

        """
        # check made for backward compatability with old topic subscribe  error response behaviour when an error
        # was thrown if the topic didnt exists on rosmaster
        if fail_on_topic_inexistence:
            topics = self.topics()
            if topic not in topics:
                return {'subscribed_error': [{topic: str.format("Unknown topic type: {0}", topic)}]}

        url = self._device_api_host + DEVICE_TOPIC_API_PATH + self.uuid + TOPIC_SUBSCRIBE
        subscription_request = {
            TOPICS: [{
                'name': topic,
                'config': {
                    'qos': qos,
                    'whitelist_tag': whitelist_tag,
                    'whitelist_field': whitelist_field
                }
            }],
            KIND: str(kind).lower()
        }
        response = self._execute_api(url, HttpMethod.POST, subscription_request, retry_limit)
        subscription_status = json.loads(get_api_response_data(response)[1])
        if len(subscription_status['subscribed_error']) > 0:
            return {'subscribed_error': subscription_status['subscribed_error']}
        return {'subscribed_success': subscription_status['subscribed_success']}

    def unsubscribe_topic(self, topic, kind=TopicKind.METRIC, retry_limit=0):

        """
        To unsubscribe a topic

        :param topic: topic name to unsubscribe
        :type topic: string
        :param kind: unsubscribe the topic as a metric or log
        :type kind: enum :py:class:`~TopicKind`
        :param retry_limit: Optional parameter to specify the number of retry attempts to be
              carried out if any failures occurs during the API call.
        :type retry_limit: int
        :return: list containing unsubscription status.
        :raises: :py:class:`~utils.error. DeviceNotFountException`: If the device is not found.
        :raises: :py:class:`~utils.error.APIError`: If the api returns an error, the status code is
            anything other than 200/201

        Following example demonstrates how to unsubscribe a topic

         >>> from rapyuta_io import Client
         >>> client = Client(auth_token='auth_token', project="project_guid")
         >>> device = client.get_device('test_device_id')
         >>> device.unsubscribe_topic('test_topic', TopicKind.METRIC)

        """

        url = self._device_api_host + DEVICE_TOPIC_API_PATH + self.uuid + TOPIC_UNSUBSCRIBE
        un_subscription_request = {TOPICS: [topic], KIND: str(kind).lower()}
        response = self._execute_api(url, HttpMethod.POST, un_subscription_request, retry_limit)
        un_subscription_status = json.loads(get_api_response_data(response)[1])
        if len(un_subscription_status['unsubscribed_error']) > 0:
            return {'unsubscribed_error': un_subscription_status['unsubscribed_error']}
        return {'unsubscribed_success': un_subscription_status['unsubscribed_success']}

    def subscribe_metric(self, metric, qos=QoS.LOW, retry_limit=0):
        """
     .. deprecated:: 0.38.0
        Will be removed in the future.

        Subscribe to given metric on the device

        :param metric: metric to be subscribed
        :type metric: SystemMetric
        :param qos: QoS value of associated with metric
        :type qos: QoS
        :param retry_limit: Optional parameter to specify the number of retry attempts to be carried
                            out if any failure occurs during API invocation
        :type retry_limit: int
        :return: returns the boolean indicating status of the operation
        :raises: :py:class:`~utils.error.DeviceNotFoundException`: If the device is not found
        :raises: :py:class:`~utils.error.APIError`: If the api returns an error, the status code is anything
            other than 200/201

        Following example demonstrate how to subscribe to a metric
          >>> from rapyuta_io import Client
          >>> from rapyuta_io.clients.device import SystemMetric, QoS
          >>> client = Client(auth_token='auth_token', project='project_guid')
          >>> device = client.get_device('device_id')
          >>> device.subscribe_metric(SystemMetric.CPU, QoS.LOW)

        """
        url = self._device_api_host + DEVICE_METRIC_API_PATH + self.uuid
        tags = {'qos': qos.value}
        subcribe_request = {
            'name': metric.value,
            'config': tags
        }
        response = self._execute_api(url, HttpMethod.POST, subcribe_request, retry_limit)
        status = get_api_response_data(response, True)
        return status.get('status') == 'success'

    def unsubscribe_metric(self, metric, retry_limit=0):
        """
     .. deprecated:: 0.38.0
        Will be removed in the future.

        Unsubscribe to given metric

        :param metric: metric to be unsubscribed
        :type metric: SystemMetric
        :param retry_limit: Optional parameter to specify the number of retry attempts to be carried
                            out if any failure occurs during API invocation
        :type retry_limit: int
        :return: boolean indicates the status of the operation
        :raises: :py:class:`~utils.error.DeviceNotFoundException`: If the device is not found
        :raises: :py:class:`~utils.error.APIError`: If the api returns an error, the status code is anything
            other than 200/201

        Following example demonstrate how to unsubscribe to a metric
         >>> from rapyuta_io import Client
         >>> from rapyuta_io.clients.device import SystemMetric
         >>> client = Client(auth_token='token', project='project_id')
         >>> device = client.get_device('device_id')
         >>> device.unsubscribe_metric(SystemMetric.CPU)

        """
        url = self._device_api_host + DEVICE_METRIC_API_PATH + self.uuid + '/' + metric.value
        response = self._execute_api(url, HttpMethod.DELETE, retry_limit=retry_limit)
        status = get_api_response_data(response, True)
        return status.get('status') == 'success'

    def metrics(self, retry_limit=0):
        """
     .. deprecated:: 0.38.0
        Will be removed in the future.

        Gets the metric status from the device

        :param retry_limit: Optional parameter to specify the number of retry attempts to be carried
                            out if any failure occurs during API invocation
        :type retry_limit: int
        :return: list of instances of class :py:class:`Metric`:
        :raises: :py:class:`~utils.error.DeviceNotFoundException`: If the device is not found
        :raises: :py:class:`~utils.error.APIError`: If the api returns an error, the status code is anything
            other than 200/201

        Following example demonstrate how to get the status of metric from the device
         >>> from rapyuta_io import Client
         >>> client = Client(auth_token='token', project='project_id')
         >>> device = client.get_device('device_id')
         >>> device.metrics()

        """
        url = self._device_api_host + DEVICE_METRIC_API_PATH + self.uuid
        response = self._execute_api(url, HttpMethod.GET, retry_limit=retry_limit)
        metrics = get_api_response_data(response)
        return [Metric(metric) for metric in metrics]

    def upload_log_file(self, upload_request, retry_limit=0):
        """
        Uploads the specified logfile in request to the cloud storage

        :param upload_request: upload specific details
        :type  upload_request: :py:class:`~.LogsUploadRequest`
        :param retry_limit: Optional parameter to specify the number of retry attempts to be carried
                            out if any failure occurs during API invocation
        :type retry_limit: int
        :return: UUID of the upload request using which user can query the status/download the file from cloud
        :raises: :py:class:`~utils.error.APIError`: If the api returns an error, the status code is anything
            other than 200/201

        Following example demonstrate how to upload the log file from device
         >>> from rapyuta_io import Client
         >>> from rapyuta_io.clients import LogsUploadRequest
         >>> client = Client(auth_token='token', project='project_id')
         >>> device = client.get_device('device_id')
         >>> upload_request = LogsUploadRequest('/path/to/file.log', file_name='new-file-name.log', override=False,
         >>>                    purge_after=False, metadata={'key': 'value'})
         >>> request_uuid = device.upload_log_file(upload_request)
        """
        url = self._device_api_host + DEVICE_LOG_API_PATH + '{}/{}'.format(self.uuid, 'upload')
        response = self._execute_api(url, HttpMethod.POST, upload_request.to_json(),
                                     retry_limit=retry_limit)
        data = get_api_response_data(response)
        return data['request_uuid']

    def get_log_upload_status(self, request_uuid, retry_limit=0):
        """
        Gets the current upload status

        :param request_uuid: UUID of the upload request
        :type request_uuid: str
        :param retry_limit: Optional parameter to specify the number of retry attempts to be carried
                            out if any failure occurs during API invocation
        :type retry_limit: int
        :return: returns instance of the class :py:class:`~.LogUploadStatus`
        :raises: :py:class:`~utils.error.LogsUUIDNotFoundException`: If the request_uuid is not found
        :raises: :py:class:`~utils.error.APIError`: If the api returns an error, the status code is anything
            other than 200/201

        Following example demonstrate how to get the status of ongoing upload operation
         >>> from rapyuta_io import Client
         >>> client = Client(auth_token='token', project='project_id')
         >>> device = client.get_device('device_id')
         >>> upload_status = device.get_log_upload_status('request_uuid')
        """
        url = self._device_api_host + DEVICE_LOG_API_PATH + '{}/status/{}'.format(self.uuid, request_uuid)
        response = self._execute_api(url, HttpMethod.GET, retry_limit=retry_limit)
        if response.status_code == requests.codes.NOT_FOUND:
            raise LogsUUIDNotFoundException('not able to find requested UUID {}'.format(request_uuid))
        data = get_api_response_data(response)
        result = LogUploadStatus(data)
        if 'shared_urls' in result:
            for shared_url in result['shared_urls']:
                setattr(shared_url, '_device_api_host', self._device_api_host)
        return result

    def cancel_log_file_upload(self, request_uuid, retry_limit=0):
        """
        Cancels the ongoing upload operation.

        :param request_uuid: UUID of the upload request
        :type request_uuid: str
        :param retry_limit: Optional parameter to specify the number of retry attempts to be carried
                                out if any failure occurs during API invocation
        :type retry_limit: int
        :return: status of the cancel operation, True indicates operation succeeded False otherwise
        :raises: :py:class:`~utils.error.LogsUUIDNotFoundException`: If the request_uuid is not found
        :raises: :py:class:`~utils.error.BadRequestError`: if the upload operation is completed
        :raises: :py:class:`~utils.error.APIError`: If the api returns an error, the status code is anything
            other than 200/201

        Following example demonstrate how to cancel the ongoing upload operation
         >>> from rapyuta_io import Client
         >>> client = Client(auth_token='token', project='project_id')
         >>> device = client.get_device('device_id')
         >>> cancel_status = device.cancel_log_file_upload('request_uuid')
        """
        url = self._device_api_host + DEVICE_LOG_API_PATH + '{}/{}/cancel'.format(self.uuid, request_uuid)
        response = self._execute_api(url, HttpMethod.PUT, retry_limit=retry_limit)
        if response.status_code == requests.codes.NOT_FOUND:
            raise LogsUUIDNotFoundException('not able to find requested UUID {}'.format(request_uuid))
        data = get_api_response_data(response)
        return data['success']

    def delete_uploaded_log_file(self, request_uuid, retry_limit=0):
        """
         Deletes the logs file from cloud storage

        :param request_uuid: UUID of the upload request
        :type request_uuid: str
        :param retry_limit: Optional parameter to specify the number of retry attempts to be carried
                           out if any failure occurs during API invocation
        :type retry_limit: int
        :return: status of the operation, True indicates operation succeeded false otherwise
        :raises: :py:class:`~utils.error.LogsUUIDNotFoundException`: If the request_uuid is not found
        :raises: :py:class:`~utils.error.BadRequestError`: if the upload is in progress
        :raises: :py:class:`~utils.error.APIError`: If the api returns an error, the status code is anything
           other than 200/201

        Following example demonstrate how to delete the uploaded log file from cloud storage
         >>> from rapyuta_io import Client
         >>> client = Client(auth_token='token', project='project_id')
         >>> device = client.get_device('device_id')
         >>> device.delete_uploaded_log_file('request_uuid')
        """
        url = self._device_api_host + DEVICE_LOG_API_PATH + '{}/{}'.format(self.uuid, request_uuid)
        response = self._execute_api(url, HttpMethod.DELETE, retry_limit=retry_limit)
        if response.status_code == requests.codes.NOT_FOUND:
            raise LogsUUIDNotFoundException('not able to find requested UUID {}'.format(request_uuid))
        data = get_api_response_data(response)
        return data['success']

    def download_log_file(self, request_uuid, retry_limit=0):
        """
        Returns downloadable log file URL, using this URL user can download log file using Wget or curl

       :param request_uuid: UUID of the upload request
       :type request_uuid: str
       :param retry_limit: Optional parameter to specify the number of retry attempts to be carried
                           out if any failure occurs during API invocation
       :type retry_limit: int
       :return: url of the uploaded log file in the cloud
       :raises: :py:class:`~utils.error.LogsUUIDNotFoundException`: If the request_uuid is not found
       :raises: :py:class:`~utils.error.BadRequestError`: if the upload operation is in progress
       :raises: :py:class:`~utils.error.APIError`: If the api returns an error, the status code is anything
           other than 200/201

       Following example demonstrate how to get the downloadable url
        >>> from rapyuta_io import Client
        >>> client = Client(auth_token='token', project='project_id')
        >>> device = client.get_device('device_id')
        >>> download_url = device.download_log_file('request_uuid')
        """
        url = self._device_api_host + DEVICE_LOG_API_PATH + '{}/{}'.format(self.uuid, request_uuid)
        response = self._execute_api(url, HttpMethod.GET, retry_limit=retry_limit)
        if response.status_code == requests.codes.NOT_FOUND:
            raise LogsUUIDNotFoundException('not able to find requested UUID {}'.format(request_uuid))
        data = get_api_response_data(response)
        return data['signed_url']

    def list_uploaded_files_for_device(self, retry_limit=0, sort=None, reverse=False, paginate=False, page_size=10,
                                       page_number=1, filter_by_filename=None, filter_by_status=None):
        """
        Returns list of the uploaded files with status and other properties

       :param retry_limit: Optional parameter to specify the number of retry attempts to be carried
                           out if any failure occurs during API invocation
       :type retry_limit: int
       :param sort: Optional parameter to sort the list of log files according to their attributes
                    (options: [filename, status, total_size, created_at, updated_at])
       :type sort: str
       :param reverse: Optional parameter to sort the list of log files in descending order
                       (ascending order by default)
       :type reverse: bool
       :param paginate: Optional parameter to paginate the list of log files
       :type paginate: bool
       :param page_size: Optional parameter to specify the number of logs to be displayed in a page.
                         (default:10, options:[10,20,30,40])
       :type page_size: int
       :param page_number: Optional parameter to specify the page number whose logs are to be displayed.
                           (default:1)
       :type page_number: int
       :param filter_by_filename: Optional parameter to filter the list of log files if the provided name
                        is a substring of any filename.
       :type filter_by_filename: str
       :param filter_by_status: Optional parameter to filter the list of log files based on their status.
                           (options : [COMPLETED, IN PROGRESS, FAILED, CANCELLED])
       :type filter_by_status: list
       :return: returns the list of instance of the class :py:class:`~.LogUploads`
       :raises: :py:class:`~utils.error.BadRequestError`: if the upload operation is in progress
       :raises: :py:class:`~utils.error.APIError`: If the api returns an error, the status code is anything
           other than 200/201

       Following example demonstrate how to get the downloadable url
        >>> from rapyuta_io import Client
        >>> client = Client(auth_token='token', project='project_id')
        >>> device = client.get_device('device_id')
        >>> status_list = device.list_uploaded_files_for_device()
        """
        url = self._device_api_host + DEVICE_LOG_API_PATH + '{}/list'.format(self.uuid)
        query_dict = self.list_log_files_query_dict(sort, reverse,
                                                    paginate, page_size, page_number,
                                                    filter_by_filename, filter_by_status)
        if query_dict:
            url += '?' + urlencode(query_dict)
        response = self._execute_api(url, HttpMethod.GET, retry_limit=retry_limit)
        data = get_api_response_data(response)
        return [LogUploads(status) for status in data]

    def list_log_files_query_dict(self, sort, reverse, paginate, page_size, page_number, filter_by_filename,
                                  filter_by_status):
        query_dict = {}
        valid_sort_list = ['status', 'filename', 'created_at', 'updated_at', 'total_size']
        if sort:
            if sort not in valid_sort_list:
                raise InvalidParameterException('sorting can only be done on the basis of {}'
                                                .format(', '.join(valid_sort_list)))
            query_dict["sort"] = sort
            if reverse:
                query_dict["sort"] = '-' + sort
        if paginate:
            page_dict = {"size": page_size, "number": page_number}
            query_dict["page"] = json.dumps(page_dict)
        filter_dict = {"and": []}
        if filter_by_filename:
            filter_dict["and"].append({"filename": {"like": "%" + filter_by_filename + "%"}})
        if filter_by_status:
            query_status_dict = {"or": []}
            for status in filter_by_status:
                query_status_dict["or"].append({"status": status})
            filter_dict["and"].append(query_status_dict)
        if filter_by_status or filter_by_filename:
            query_dict["filter"] = json.dumps(filter_dict)
        return query_dict

    def create_shared_url(self, shared_url):
        """
        Create a SharedURL that let's you download the Log file from a Public URL.

       :param shared_url: Instance of the SharedURL Object
       :type shared_url: :py:class:`~rapyuta_io.clients.model.SharedURL`
       :return: Instance of :py:class:`~rapyuta_io.clients.model.SharedURL` class

       Following example demonstrate how to get the downloadable url
        >>> from datetime import timedelta, datetime
        >>> from rapyuta_io import Client
        >>> expiry_time = datetime.now() + timedelta(days=7)
        >>> client = Client(auth_token='token', project='project_id')
        >>> device = client.get_device('device_id')
        >>> log = [log.filename == 'file.log' for log in device.list_uploaded_files_for_device()][0]
        >>> shared_url = device.create_shared_url(SharedURL(log.request_uuid, expiry_time=expiry_time))
        >>> print shared_url.url
        """
        url = self._device_api_host + DEVICE_LOG_API_PATH + '{}/shared_urls'.format(shared_url.request_uuid)
        response = self._execute_api(url, HttpMethod.POST, payload=shared_url.serialize())
        if response.status_code == requests.codes.NOT_FOUND:
            raise LogsUUIDNotFoundException('not able to find requested UUID {}'.format(shared_url.request_uuid))
        data = get_api_response_data(response)
        shared_url = SharedURL.deserialize(data)
        setattr(shared_url, '_device_api_host', self._device_api_host)
        return shared_url

    def upgrade(self):
        """
        Upgrade a device running with python2 to python3

        :raises: :py:class:`~utils.error.DeviceNotFoundException`:  If the device is not found.
        :raises: :py:class:`~utils.error.DeploymentRunningException`:  When any deployment is
                 running on the device.
        :raises: :py:class:`~utils.error.APIError`: If the api returns an error, the status code
                  is anything other than 200/201

        Following example demonstrate how to upgrade a device onboarded on python2 runtime

        >>> from rapyuta_io import Client
        >>> client = Client(auth_token='token', project='project_id')
        >>> device = client.get_device('device_id')
        >>> device.upgrade()
        >>> onboard_script = device.onboard_script()
        >>> onboard_script.run()
        """
        for deployment in self.deployments:
            if deployment.phase.lower() == DeploymentPhaseConstants.SUCCEEDED.value.lower():
                raise DeploymentRunningException('Cannot upgrade device. Deployment %s is running '
                                                 'on the device' % deployment.deployment_id)

        url = self._device_api_host + DEVICE_API_PATH + self.uuid + '/upgrade'
        response = self._execute_api(url, HttpMethod.PUT)
        if response.status_code == requests.codes.BAD_REQUEST:
            raise DeploymentRunningException()
        get_api_response_data(response)


class ROSDistro(str, enum.Enum):
    """
    Enumeration variables for the Supported ROS Distros. ROS Distro may be one of: \n
    ROSDistro.KINETIC ('kinetic') \n
    ROSDistro.MELODIC ('melodic') \n
    ROSDistro.NOETIC ('noetic') \n
    """

    def __str__(self):
        return str(self.value)

    KINETIC = 'kinetic'
    MELODIC = 'melodic'
    NOETIC = 'noetic'


class DeploymentPhaseConstants(str, enum.Enum):
    """
    Enumeration variables for the deployment phase

    Deployment phase can be any of the following types \n
    DeploymentPhaseConstants.INPROGRESS \n
    DeploymentPhaseConstants.PROVISIONING \n
    DeploymentPhaseConstants.SUCCEEDED \n
    DeploymentPhaseConstants.FAILED_TO_START \n
    DeploymentPhaseConstants.PARTIALLY_DEPROVISIONED \n
    DeploymentPhaseConstants.DEPLOYMENT_STOPPED \n
    """

    def __str__(self):
        return str(self.value)

    INPROGRESS = 'In progress'
    PROVISIONING = 'Provisioning'
    SUCCEEDED = 'Succeeded'
    FAILED_TO_START = 'Failed to start'
    PARTIALLY_DEPROVISIONED = 'Partially deprovisioned'
    DEPLOYMENT_STOPPED = 'Deployment stopped'
